import copy
import logging
import os
import questionary
import sys

import esm_parser
import yaml
import esm_utilities
from esm_calendar import Calendar, Date
from loguru import logger

from . import batch_system, helpers


def run_job(config):
    helpers.evaluate(config, "prepare", "prepare_recipe")
    return config


def mini_resolve_variable_date_file(date_file, config):
    while "${" in date_file:
        pre, post = date_file.split("${", 1)
        variable, post = post.split("}", 1)
        if "." in variable:
            variable_section, variable = variable.split(".")
            answer = config[variable_section].get(variable)
        else:
            answer = config["general"].get(variable)
            if not answer:
                answer = config.get("env", {}).get(variable)
                if not answer:
                    try:
                        assert variable.startswith("env.") or variable.startswith(
                            "general."
                        )
                    except AssertionError:
                        print(
                            "The date file contains a variable which is not in the >>env<< or >>general<< section. This is not allowed!"
                        )
                        print(f"date_file = {date_file}")
                        sys.exit(1)
        date_file = pre + answer + post
    return date_file


def _read_date_file(config):
    import logging
    import os

    date_file = (
        f"{config['general']['experiment_dir']}/scripts"
        f"/{config['general']['expid']}_{config['general']['setup_name']}.date"
    )

    date_file = mini_resolve_variable_date_file(date_file, config)

    if os.path.isfile(date_file):
        logging.info("Date file read from %s", date_file)
        with open(date_file) as date_file:
            date, run_number = date_file.readline().strip().split()
            run_number = int(run_number)
        write_file = False
    else:
        logging.info("No date file found %s", date_file)
        logging.info("Initializing run_number=1 and date=18500101")
        date = config["general"].get("initial_date", "18500101")
        run_number = 1
        write_file = True
    config["general"]["run_number"] = run_number
    config["general"]["current_date"] = date
    logging.info("current_date = %s", date)
    logging.info("run_number = %s", run_number)
    return config


def check_model_lresume(config):
    if config["general"]["run_number"] != 1:
        for model in config["general"]["valid_model_names"]:
            config[model]["lresume"] = True
    else:
        # Did the user give a value? If yes, keep it, if not, first run:
        for model in config["general"]["valid_model_names"]:
            if "lresume" in config[model]:
                user_lresume = config[model]["lresume"]
            else:
                user_lresume = False

            if isinstance(user_lresume, str) and "${" in user_lresume:
                user_lresume = esm_parser.find_variable(
                    model, user_lresume, config, [], []
                )
            if type(user_lresume) == str:

                if user_lresume == "0" or user_lresume.upper() == "FALSE":
                    user_lresume = False
                elif user_lresume == "1" or user_lresume.upper() == "TRUE":
                    user_lresume = True
            elif isinstance(user_lresume, int):
                if user_lresume == 0:
                    user_lresume = False
                elif user_lresume == 1:
                    user_lresume = True
            config[model]["lresume"] = user_lresume
    for model in config["general"]["valid_model_names"]:
        # Check if lresume contains a variable which might be set in a different model, and resolve this case
        if (
            "lresume" in config[model]
            and isinstance(config[model]["lresume"], str)
            and "${" in config[model]["lresume"]
        ):
            lr = esm_parser.find_variable(
                model, config[model]["lresume"], config, [], []
            )
            config[model]["lresume"] = eval(lr)
    return config


def resolve_some_choose_blocks(config):
    # from esm_parser import choose_blocks

    # Component-specific environment variables into ``computer``
    # before ``computer`` ``choose_`` blocks are resolved
    model_env_into_computer(config)

    esm_parser.choose_blocks(config, blackdict=config._blackdict)
    return config


def model_env_into_computer(config):
    """
    This function allows to store in the ``computer`` dictionary, variables that were
    defined inside ``environment_changes`` or ``compile/runtime_environment_changes``
    in the components.

    It excludes ``module_actions`` and ``export_vars`` dictionaries as those are
    resolved later.

    This function is necessary for controlling ``choose_`` blocks in the computer file
    from the component configuration file (i.e. add ``useMPI`` case to the component to
    control which ``useMPI`` case is selected in the computer file).

    This function works both for compilation time and run time, and the result is that
    all components work under the same environment. The only exception is for
    the compilation of components, where ``add_export_vars`` and ``add_module_actions``
    are excluded from the merging into ``computer``, and are included individually in
    respective compilation scripts.

    Later on, it might be desirable to always split the environments both for compiling
    (done by Paul Gierz, but this function would need to be adapted) and running (not
    done yet).

    If this script gives you problems contact Miguel Andres-Martinez
    (miguel.andres-martinez@awi.de).

    Parameters
    ----------
    config : dict
        Dictionary containing the simulation/compilation information

    Raises
    ------
    User Note/Error
        If the same variable is found in two or more different component environments.
        Asks the user how to proceed.
    """

    # import logging
    # from esm_parser import basic_choose_blocks, dict_merge, user_note, user_error, pprint_config

    # Get which type of changes are to be applied to the environment
    run_or_compile = config.get("general", {}).get("run_or_compile", "runtime")
    config["general"]["run_or_compile"] = run_or_compile
    thesechanges = run_or_compile + "_environment_changes"
    # List the component names
    models = config.get("general", {}).get("models", [])
    # ``env_vars`` stores information about the environment changes to inform
    # about conflicts
    env_vars = {}
    # Loop through the models
    for model in models:
        # Update the models ``environment_changes`` with the ``compiletime/runtime_
        # environment_changes
        modelconfig = copy.deepcopy(config[model])
        modelconfig["environment_changes"] = modelconfig.get("environment_changes", {})
        if thesechanges in modelconfig:
            if "environment_changes" in modelconfig:
                modelconfig["environment_changes"].update(modelconfig[thesechanges])
            else:
                modelconfig["environment_changes"] = modelconfig[thesechanges]
        # Resolve ``choose_`` blocks, ``add_`` and ``remove_`` inside ``environment_
        # changes``
        esm_parser.basic_choose_blocks(modelconfig["environment_changes"], config)
        # Set to true when specified by the user in ``env_overwrite`` or when this
        # method has been already called once in this run
        overwrite = config[model].get("env_overwrite", False)
        # Loop through every variable in ``environment_changes`` except ``export_vars``
        # and ``module_actions`` as those are solved by ``esm_environment`` later and
        # need of the solving of later ``choose_`` blocks.
        for key, value in modelconfig["environment_changes"].items():
            if (
                key
                not in [
                    "export_vars",
                    "module_actions",
                    "add_export_vars",
                    "add_module_actions",
                    "unset_vars",
                    "add_unset_vars",
                ]
                and "computer" in config
                and not overwrite
                # and run_or_compile=="runtime"
            ):
                # If the key is already included in ``env_vars``, the key variable has
                # been already modified by a previous model and a warning needs to be
                # raised. Do this check only on the first step (user interaction step).
                if key in env_vars and config["general"]["run_number"] == 1:
                    # Previous model with the same key
                    model0 = env_vars[key][1]
                    while True:
                        # Warn the user about the overwriting of the variable
                        esm_parser.user_note("Environment conflict", f"In '{model0}':")
                        esm_parser.pprint_config({key: env_vars[key][0]})
                        logging.info("\nIn '" + model + "':")
                        esm_parser.pprint_config({key: value})
                        # Ask the user how to proceed if it is not a `tidy job
                        if not config["general"]["jobtype"] == "tidy":
                            user_answer = input(
                                f"Environment variable '{key}' defined in '{model0}' is "
                                + "going to be overwritten by the one defined in "
                                + f"'{model}'. Are you okay with that? (Y/n): "
                            )
                        # If it is a ``tidy`` job, the user has already
                        # interacted and accepted the overwriting.
                        else:
                            user_answer = "Y"
                            logging.info(
                                f"Environment variable '{key}' defined in '{model0}' is "
                                + f"overwritten by the one defined in '{model}'."
                            )
                        # If the user selects ``Y``, overwrite the environment variable
                        if user_answer == "Y":
                            config[model]["env_overwrite"] = True
                            break
                        # If the user selects ``n`` raise a user error with recommendations
                        elif user_answer == "n":
                            config[model]["env_overwrite"] = False
                            esm_parser.user_error(
                                "Environment conflict",
                                "You were not happy with the environment variable "
                                + f"'{key}' in '{model0}' being overwritten by the same "
                                + f"variable in '{model}'. If you are running a "
                                + "coupled setup, we recommend that you resolve this "
                                + "conflict inside the coupled setup file, by "
                                + "specifying unconflicting environments for each "
                                + "model, or contact the setup developers.",
                            )
                        else:
                            logging.info("Wrong answer, please choose Y/n.")
                # Merge variable into the ``computer`` dictionary so that it becomes
                # part of the general environment.
                esm_parser.dict_merge(config["computer"], {key: value})
                # Add the variable to ``env_vars`` so it can be checked for conflicts
                # with other models.
                env_vars[key] = [value, model]


def _initialize_calendar(config):
    config = set_restart_chunk(config)
    config = set_leapyear(config)
    config = set_overall_calendar(config)
    if config["general"]["reset_calendar_to_last"]:
        config = find_last_prepared_run(config)
    config = set_most_dates(config)
    if not "iterative_coupling" in config["general"]:
        config["general"]["chunk_number"] = 1

        if config["general"]["run_number"] == 1:
            config["general"]["first_run_in_chunk"] = True
        else:
            config["general"]["first_run_in_chunk"] = False
        if config["general"]["next_date"] >= config["general"]["final_date"]:
            config["general"]["last_run_in_chunk"] = True
        else:
            config["general"]["last_run_in_chunk"] = False
    return config


def set_restart_chunk(config):
    nyear, nmonth, nday, nhour, nminute, nsecond = 0, 0, 0, 0, 0, 0
    nyear = int(config["general"].get("nyear", nyear))
    if not nyear:
        nmonth = int(config["general"].get("nmonth", nmonth))
    if not nyear and not nmonth:
        nday = int(config["general"].get("nday", nday))
    if not nyear and not nmonth and not nday:
        nhour = int(config["general"].get("nhour", nhour))
    if not nyear and not nmonth and not nday and not nhour:
        nminute = int(config["general"].get("nminute", nminute))
    if not nyear and not nmonth and not nday and not nhour and not nminute:
        nsecond = int(config["general"].get("nsecond", nsecond))
    if (
        not nyear
        and not nmonth
        and not nday
        and not nhour
        and not nminute
        and not nsecond
    ):
        nyear = 1
    config["general"]["nyear"] = nyear
    config["general"]["nmonth"] = nmonth
    config["general"]["nday"] = nday
    config["general"]["nhour"] = nhour
    config["general"]["nminute"] = nminute
    config["general"]["nsecond"] = nsecond
    return config


def set_leapyear(config):
    # make sure all models agree on leapyear
    if "leapyear" in config["general"]:
        for model in config["general"]["valid_model_names"]:
            config[model]["leapyear"] = config["general"]["leapyear"]
    else:
        for model in config["general"]["valid_model_names"]:
            if "leapyear" in config[model]:
                for other_model in config["general"]["valid_model_names"]:
                    if "leapyear" in config[other_model]:
                        if (
                            not config[other_model]["leapyear"]
                            == config[model]["leapyear"]
                        ):
                            print(
                                "Models "
                                + model
                                + " and "
                                + other_model
                                + " do not agree on leapyear. Stopping."
                            )
                            sys.exit(43)
                    else:
                        config[other_model]["leapyear"] = config[model]["leapyear"]
                config["general"]["leapyear"] = config[model]["leapyear"]
                break

    if not "leapyear" in config["general"]:
        for model in config["general"]["valid_model_names"]:
            config[model]["leapyear"] = True
        config["general"]["leapyear"] = True
    return config


def set_overall_calendar(config):

    # set the overall calendar
    if config["general"]["leapyear"]:
        config["general"]["calendar"] = Calendar(1)
    else:
        config["general"]["calendar"] = Calendar(0)
    return config


def find_last_prepared_run(config):

    calendar = config["general"]["calendar"]
    current_date = Date(config["general"]["current_date"], calendar)
    initial_date = Date(config["general"]["initial_date"], calendar)
    delta_date = (
        config["general"]["nyear"],
        config["general"]["nmonth"],
        config["general"]["nday"],
        config["general"]["nhour"],
        config["general"]["nminute"],
        config["general"]["nsecond"],
    )

    while True:
        if current_date < initial_date:
            break

        next_date = current_date.add(delta_date)
        end_date = next_date - (0, 0, 1, 0, 0, 0)

        datestamp = (
            current_date.format(form=9, givenph=False, givenpm=False, givenps=False)
            + "-"
            + end_date.format(form=9, givenph=False, givenpm=False, givenps=False)
        )

        # Solve base_dir with variables
        base_dir = esm_parser.find_variable(
            ["general", "base_dir"], config["general"]["base_dir"], config, [], True
        )
        expid = config["general"]["expid"]
        it_coupled_model_name = config["general"]["iterative_coupled_model"]

        if os.path.isdir(
            f"{base_dir}/{expid}/run_{it_coupled_model_name}{datestamp}"
        ):
            config["general"]["current_date"] = current_date
            return config

        current_date = current_date - delta_date

    print("ERROR: Could not find a prepared run.")
    sys.exit(42)


def set_most_dates(config):

    calendar = config["general"]["calendar"]
    if isinstance(config["general"]["current_date"], Date):
        current_date = config["general"]["current_date"]
    else:
        current_date = Date(config["general"]["current_date"], calendar)
    delta_date = (
        config["general"]["nyear"],
        config["general"]["nmonth"],
        config["general"]["nday"],
        config["general"]["nhour"],
        config["general"]["nminute"],
        config["general"]["nsecond"],
    )

    config["general"]["delta_date"] = delta_date
    config["general"]["current_date"] = current_date
    config["general"]["start_date"] = current_date
    config["general"]["initial_date"] = Date(
        config["general"]["initial_date"], calendar
    )
    config["general"]["final_date"] = Date(config["general"]["final_date"], calendar)
    config["general"]["prev_date"] = current_date - (0, 0, 1, 0, 0, 0)

    config["general"]["next_date"] = current_date.add(delta_date)
    config["general"]["last_start_date"] = current_date - delta_date
    config["general"]["end_date"] = config["general"]["next_date"] - (0, 0, 1, 0, 0, 0)

    config["general"]["runtime"] = (
        config["general"]["next_date"] - config["general"]["current_date"]
    )

    config["general"]["total_runtime"] = (
        config["general"]["next_date"] - config["general"]["initial_date"]
    )

    config["general"]["run_datestamp"] = (
        config["general"]["current_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
        + "-"
        + config["general"]["end_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
    )

    config["general"]["last_run_datestamp"] = (
        config["general"]["last_start_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
        + "-"
        + config["general"]["prev_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
    )
    return config


def _add_all_folders(config):
    all_filetypes = [
        "analysis",
        "config",
        "log",
        "mon",
        "couple",
        "scripts",
        "ignore",
        "unknown",
        "src",
    ]
    config["general"]["out_filetypes"] = [
        "analysis",
        "log",
        "mon",
        "scripts",
        "ignore",
        "unknown",
        "outdata",
        "restart_out",
    ]
    config["general"]["in_filetypes"] = [
        "scripts",
        "input",
        "forcing",
        "bin",
        "config",
        "restart_in",
    ]
    config["general"]["reusable_filetypes"] = config["general"].get(
        "reusable_filetypes", ["bin", "src"]
    )
    # Define the files that could be reusable accross runs (external files)
    config["general"]["potentially_reusable_filetypes"] = (
        all_filetypes + config["general"]["in_filetypes"]
    )
    # Apply changes from ``--update-files`` flag
    config = helpers.update_reusable_filetypes(config)

    experiment_dir = config["general"]["experiment_dir"]
    it_coupled_model_name = config["general"]["iterative_coupled_model"]
    datestamp = config["general"]["run_datestamp"]
    config["general"]["thisrun_dir"] = (
        f"{experiment_dir}/run_{it_coupled_model_name}{datestamp}"
    )

    for filetype in all_filetypes:
        config["general"]["experiment_" + filetype + "_dir"] = (
            config["general"]["experiment_dir"] + "/" + filetype + "/"
        )

    all_filetypes.append("work")
    config["general"]["all_filetypes"] = all_filetypes

    for filetype in all_filetypes:
        config["general"]["thisrun_" + filetype + "_dir"] = (
            config["general"]["thisrun_dir"] + "/" + filetype + "/"
        )

    config["general"]["work_dir"] = config["general"]["thisrun_work_dir"]

    all_model_filetypes = [
        "analysis",
        "bin",
        "config",
        "forcing",
        "input",
        "couple",
        "log",
        "mon",
        "outdata",
        "restart_in",
        "restart_out",
        "viz",
        "ignore",
    ]

    config["general"]["all_model_filetypes"] = all_model_filetypes

    for model in config["general"]["valid_model_names"]:
        for filetype in all_model_filetypes:
            if "restart" in filetype:
                filedir = "restart"
            else:
                filedir = filetype
            config[model]["experiment_" + filetype + "_dir"] = (
                config["general"]["experiment_dir"] + "/" + filedir + "/" + model + "/"
            )
            config[model]["thisrun_" + filetype + "_dir"] = (
                config["general"]["thisrun_dir"] + "/" + filedir + "/" + model + "/"
            )
            config[model]["all_filetypes"] = all_model_filetypes

    return config


def set_prev_date(config):
    import esm_parser

    """Sets several variables relevant for the previous date. Loops over all models in ``valid_model_names``, and sets model variables for:
    * ``prev_date``
    """
    for model in config["general"]["valid_model_names"]:
        if "time_step" in config[model] and not (
            isinstance(config[model]["time_step"], str)
            and "${" in config[model]["time_step"]
        ):
            config[model]["prev_date"] = config["general"]["current_date"] - (
                0,
                0,
                0,
                0,
                0,
                int(config[model]["time_step"]),
            )
            config[model]["last_parent_date"] = config[model]["prev_date"] - (
                0,
                0,
                0,
                0,
                0,
                int(config[model]["time_step"]),
            )

        # NOTE(PG, MAM): Here we check if the time step still has a variable which might be set in a different model, and resolve this case
        elif "time_step" in config[model] and (
            isinstance(config[model]["time_step"], str)
            and "${" in config[model]["time_step"]
        ):
            dt = esm_parser.find_variable(
                model, config[model]["time_step"], config, [], []
            )
            config[model]["prev_date"] = config["general"]["current_date"] - (
                0,
                0,
                0,
                0,
                0,
                int(dt),
            )
            config[model]["last_parent_date"] = config[model]["prev_date"] - (
                0,
                0,
                0,
                0,
                0,
                int(dt),
            )

        else:
            config[model]["prev_date"] = config["general"]["current_date"]
    return config


def set_parent_info(config):

    """Sets several variables relevant for the previous date. Loops over all models in ``valid_model_names``, and sets model variables for:
    * ``parent_expid``
    * ``parent_date``
    * ``parent_restart_dir``
    """

    # Make sure "ini_parent_dir" and "ini_restart_dir" both work:
    for model in config["general"]["valid_model_names"]:
        if not "ini_parent_dir" in config[model]:
            if "ini_restart_dir" in config[model]:
                config[model]["ini_parent_dir"] = config[model]["ini_restart_dir"]
        if not "ini_parent_exp_id" in config[model]:
            if "ini_restart_exp_id" in config[model]:
                config[model]["ini_parent_exp_id"] = config[model]["ini_restart_exp_id"]
        if not "ini_parent_date" in config[model]:
            if "ini_restart_date" in config[model]:
                config[model]["ini_parent_date"] = config[model]["ini_restart_date"]

    # check if parent is defined in esm_tools style
    # (only given for setup)
    setup = config["general"]["setup_name"]
    if not setup in config:
        setup = "general"
    if "ini_parent_exp_id" in config[setup]:
        for model in config["general"]["valid_model_names"]:
            if not "ini_parent_exp_id" in config[model]:
                config[model]["ini_parent_exp_id"] = config[setup]["ini_parent_exp_id"]
    if "ini_parent_date" in config[setup]:
        for model in config["general"]["valid_model_names"]:
            if not "ini_parent_date" in config[model]:
                config[model]["ini_parent_date"] = config[setup]["ini_parent_date"]
    if "ini_parent_dir" in config[setup]:
        for model in config["general"]["valid_model_names"]:
            if not "ini_parent_dir" in config[model]:
                config[model]["ini_parent_dir"] = (
                    config[setup]["ini_parent_dir"] + "/" + model
                )

    # Get correct parent info
    for model in config["general"]["valid_model_names"]:
        if config[model]["lresume"] == True and config["general"]["run_number"] == 1:
            config[model]["parent_expid"] = config[model]["ini_parent_exp_id"]
            if "parent_date" not in config[model]:
                config[model]["parent_date"] = config[model]["ini_parent_date"]
            config[model]["parent_restart_dir"] = config[model]["ini_parent_dir"]
        else:
            config[model]["parent_expid"] = config["general"]["expid"]
            if "parent_date" not in config[model]:
                config[model]["parent_date"] = config[model]["prev_date"]
            config[model]["parent_restart_dir"] = config[model][
                "experiment_restart_in_dir"
            ]
    return config


def add_vcs_info(config):
    """
    Adds version control system information in a plain text yaml file under the
    experiment log directory for both the models and the esm-tools.

    Parameters
    ----------
    config : dict
        The experiment configuration

    Returns
    -------
    config : dict
        The experiment configuration
    """
    exp_vcs_info_file = f"{config['general']['thisrun_log_dir']}/{config['general']['expid']}_vcs_info.yaml"
    logger.debug("Experiment information is being stored for usage under:")
    logger.debug(f">>> {exp_vcs_info_file}")
    vcs_versions = {}
    all_models = config.get("general", {}).get("models", [])
    for model in all_models:
        logger.debug(f"Locating {model}")
        try:
            model_dir = config[model]["model_dir"]
        except KeyError:
            # XIOS does not seem to define model_dir? Jan? What?
            vcs_versions[model] = f"Unable to locate model_dir for {model}."
            continue
        if helpers.is_git_repo(model_dir):
            vcs_versions[model] = helpers.get_all_git_info(model_dir)
        else:
            vcs_versions[model] = "Not a git-controlled model!"

    # NOTE(PG): There is no good way to get the repo directory from the config,
    # this may at least be a good start:
    esm_tools_repo = config.get("general", {}).get("esm_function_dir")
    if esm_tools_repo is not None:
        vcs_versions["esm_tools"] = helpers.get_all_git_info(f"{esm_tools_repo}/../")
    else:
        # FIXME(PG): This should absolutely never happen. The error message could use a better wording though...
        esm_parser.user_error("esm_tools doesn't know where it's own install location is. Something is very seriously wrong.")
    with open(exp_vcs_info_file, "w") as f:
        yaml.dump(vcs_versions, f)
    return config

def check_vcs_info_against_last_run(config):
    """
    Ensures that the version control info for two runs is identical between the
    current run and the previous run.


    This is checked by ensuring that the dictionary representation of the VCS
    files is identical. Differences will result in a ``user_error``, thus
    terminating the run. The check can be circumvented by setting::

        general:
            allow_vcs_differences: True

    Parameters
    ----------
    config : dict
        The experiment configuration

    Returns
    -------
    config : dict
        The experiment configuration
    """
    # FIXME(PG): Sometimes general.run_number is None (shows up as null in the
    # config), so this check is absolutely the worst way of doing it, but
    # whatever...
    if config['general']['run_number'] == 1 or config["general"]["run_number"] is None:
        return config  # No check needed on the very first run
    exp_vcs_info_file = f"{config['general']['thisrun_log_dir']}/{config['general']['expid']}_vcs_info.yaml"
    # FIXME(PG): This file might not exist if people erase every run folder...
    last_exp_vcs_info_file = f"{config['prev_run']['general']['thisrun_log_dir']}/{config['general']['expid']}_vcs_info.yaml"

    try:
        with open(exp_vcs_info_file, "r") as f:
            current_vcs_info = yaml.safe_load(f)
    except IOError:
        logger.warning(f"Unable to open {exp_vcs_info_file}, skipping check...")
        return config
    try:
        with open(last_exp_vcs_info_file, "r") as f:
            last_vcs_info = yaml.safe_load(f)
    except IOError:
        logger.warning(f"Unable to open {last_exp_vcs_info_file}, skipping_check...")
        return config
    if not config["general"].get("allow_vcs_differences", False) and current_vcs_info != last_vcs_info:
        esm_parser.user_error("VCS Differences", """
            You have differences in either the model code or in the esm-tools between two runs! 

            If you are **sure** that this is OK, you can set 'general.allow_vcs_differences' to True to avoid this check.
            """)


    return config


def finalize_config(config):
    config.finalize()
    return config


def add_submission_info(config):

    bs = batch_system.batch_system(config, config["computer"]["batch_system"])

    submitted = bs.check_if_submitted()
    if submitted:
        jobid = bs.get_jobid()
    else:
        jobid = os.getpid()

    config["general"]["submitted"] = submitted
    config["general"]["jobid"] = jobid
    return config


def initialize_batch_system(config):

    config["general"]["batch"] = batch_system.batch_system(
        config, config["computer"]["batch_system"]
    )

    return config


def initialize_coupler(config):
    if config["general"]["standalone"] == False:
        from . import coupler

        base_dir = config["general"]["base_dir"]
        expid = config["general"]["expid"]
        it_coupled_model_name = config["general"]["iterative_coupled_model"]
        datestamp = config["general"]["run_datestamp"]
        for model in list(config):
            if model in coupler.known_couplers:
                config["general"]["coupler_config_dir"] = (
                    f"{base_dir}/{expid}/"
                    f"run_{it_coupled_model_name}{datestamp}/config/{model}/"
                )
                config["general"]["coupler"] = coupler.coupler_class(config, model)
                break
        config["general"]["coupler"].add_files(config)
    return config


def set_logfile(config):
    log_file_path = (
        f"{config['general']['experiment_log_dir']}"
        f"/{config['general']['expid']}"
        f"_{config['general']['setup_name']}.log"
    )
    config["general"]["experiment_log_file"] = config["general"].get(
        "experiment_log_file", log_file_path
    )
    return config


def check_config_for_warnings_errors(config):
    """
    Wrapper method to call ``warn_error`` from the `ESM-Tools` recipes. Loops through
    the trigger variables to call ``warn_error``, only including ``warning`` in the
    ``triggers`` if it is an interactive session (i.e. user just submitted a job), or
    if it's not a resubmission (to avoid displaying the warning twice).

    .. note:: use this in the recipe after the ``choose_`` blocks are resolved,
       otherwise, warnings and errors nested in the ``choose_`` blocks won't be
       reached.

    Parameters
    ----------
    config : dict, esm_parser.ConfigSetup
            ConfigSetup object containing the information of the current simulation

    Returns
    -------
    config : dict, esm_parser.ConfigSetup
            ConfigSetup object containing the information of the current simulation
    """

    # Initialize the trigger variables (i.e. ``error`` and ``warning``))
    triggers = {"error": {"note_function": esm_parser.user_error}}

    # Find conditions to warn (avoid warning more than once)
    last_jobtype = config["general"].get("last_jobtype", "")
    isresubmitted = last_jobtype == config["general"]["jobtype"]
    isinteractive = config["general"].get("isinteractive", "")

    # Only warn if it is an interactive session or while submitted
    if not isresubmitted or isinteractive:
        triggers["warning"] = {"note_function": esm_parser.user_note}

    # Loop through the triggers
    for trigger, trigger_info in triggers.items():
        warn_error(config, trigger, trigger_info["note_function"])

    return config

def warn_error(config, trigger, note_function):
    """
    Checks the ``sections`` of the ``config`` for a given ``trigger`` (``"error"`` or
    ``"warning"``), and if found, returns the ``error`` or ``warning`` using
    ``note_functon`` (``user_error`` or ``user_note`` respectively). Errors always halt
    the core with ``sys.exit(1)``. Warnings only halt the code to ask the user for
    continuation if the job is interactive, it is specified inside the warning as
    ``ask_user_to_continue: True`` and the user has not called ``esm_runscripts`` with
    the ``--ignore-config-warnings`` or defined ``general.ignore_config_warnings: True``
    in their runscript.

    The syntax in the yaml files for triggering warnings or errors is as follows:
    .. code-block:: yaml

       warning/error:
           <name>: # Name for the specific warning or error
               message: "the message of the warning/error"
               esm_tools_version: ">/</=/!=/version_number" # trigger it under certain ESM-Tools version conditions
               ask_user_to_continue: True/False # Ask user about continuing or stopping the process, only for warnings, errors always kill the process

    Example
    -------
    .. code-block:: yaml

       recom:
           choose_scenario:
               HIST:
                   [ ... ]
               PI-CTRL:
                   [ ... ]
               "*":
                   add_warning:
                       "wrong scenario type":
                           message: "The scenario you specified (``${recom.scenario}``) is not supported!"
                           ask_user_to_continue: True

    If you then define ``recom.scenario: hist`` instead of ``HIST`` then you'll get the
    following:
    .. code-block::

       wrong scenario type WARNING
       ---------------------------
       Section: recom

       Wrong scenario, scenario hist does not exist

       ? Do you want to continue (set general.ignore_config_warnings: False to avoid quesitoning)?

    Parameters
    ----------
    config : dict, esm_parser.ConfigSetup
        ConfigSetup object containing the information of the current simulation
    trigger : string
        ``error`` or ``warning``
    note_function : esm_parser.user_error, esm_parser.user_note
        Method to report the note
    """
    # Sufixes for the warning special case
    if trigger=="warning":
        sufix_name = f" WARNING"
    else:
        sufix_name = f""

    # Loop through the sections (e.g. ``general``, ``<model>``, ``computer``, ...)
    for section, value in config.items():
        # If the trigger is found output the warnings or errors
        if trigger in value:
            actions = value[trigger]
            # Loop through the warnings or errors
            for action_name, action_info in actions.items():
                # Check if the version condition for the error/warning is met
                version_condition = action_info.get("esm_tools_version", ">0.0.0")
                if esm_utilities.check_valid_version(version_condition):
                    # Call the ``note_function`` (``user_error`` for errors,
                    # ``user_note`` for warnings))
                    note_function(
                        f"{action_name}{sufix_name}",
                        f'Section: ``{section}``\n\n{action_info.get("message", "")}'
                    )

                    # Check if the warning should halt execution and ask the user if
                    # if is an interactive session, the warning info specifies that i
                    # needs to halt, and the user has not defined the
                    # ``--ignore-config-warnings`` flag in the ``esm_runscripts`` call
                    if (
                        trigger=="warning"
                        and config["general"].get("isinteractive")
                        and action_info.get("ask_user_to_continue", False)
                        and not config["general"].get("ignore_config_warnings", False)
                    ):
                        proceed = questionary.confirm(
                            "Do you want to continue (set "
                            "general.ignore_config_warnings: False to avoid "
                            "quesitoning)?"
                        ).ask()
                        if not proceed:
                            sys.exit(1)



import copy
import glob
import os
import re
import shutil
import time
import yaml
from esm_parser import determine_computer_from_hostname
from loguru import logger

from .output import *
from .read_shipped_data import *
from .test_utilities import *

# Bold strings
bs = "\033[1m"
be = "\033[0m"

# Define default files for comparisson
compare_files = {"comp": ["comp-"], "run": [".run", "finished_config", "namelists"]}

"""
``config.yaml``:
    This file includes exceptions rules to be applied during the tests (such as copying
    instead of downloading the source code for compilation), and output and restart
    files to be checked that they exist after an actual run. This file needs to exist
    in each ``resources/runscripts/<model>`` folder.
"""

# Coded by @pgierz, but not finished, therefore commented.
# class Comparison:
#    """Compares two ESM Tools files"""
#
#    def __init__(self, test_file, truth_file):
#        """
#        Parameters
#        ----------
#        test_file : str
#            str representation (already opened and read in) of the file (e.g.
#            run file, compilatoin script, namelist) which you want to check
#        truth_file : str
#            str representation of the file which is known to be a valid truth.
#        """
#        self.test_file = test_file
#        self.truth_file = truth_file
#
#    @classmethod
#    def from_filepaths(cls, test_file, truth_file):
#        with open(test_file, "r") as f1:
#            test_file = f1.read()
#        with open(test_file, "r") as f2:
#            truth_file = f2.read()
#        return cls(test_file, truth_file)
#
#    @classmethod
#    def from_test_filepath_and_pkg(cls, test_file, truth_file):
#        """
#        Given a test filepath, and a relative truth path, return a new Comparison.
#
#        For the relative truth path, we are reading from the package. So,
#        assuming you want to use the following as your truth:
#
#        >>> truth_file = "ollie/run/awicm/awicm2-initial-monthly/scripts/awicm2-initial-monthly_compute_20000101 -20000131.run"
#
#        This would check the run file for a run of awicm using a awicm2
#        initialization with monthly restarts which is run on the ollie HPC.
#        """
#        # get last tested (Truth)
#        with open(test_file, "r") as f:
#            test_file = f1.read()
#        truth_file = read_shipped_data.get_last_tested(truth_file)
#        return cls(test_file, truth_file)
#
#
# class CompileFileComparison(Comparison):
#    pass
#
#
# class SadFileComparison(Comparison):
#    pass
#
#
# class FinishedESMConfigComparison(Comparison):
#    pass
#
#
# def NamelistsComparison(Comparison):
#    pass


#######################################################################################
# TESTS
#######################################################################################
def comp_test(info):
    """
    Downloads/copies the source codes of the models to be tested and compiles them. If
    the ``--check`` flag is provided in the ``esm_tests`` call, it does not
    download/copy and compile the source codes, but produces the ``comp-*.sh``
    compilation scripts instead, for comparison purposes with last tested versions of
    them.

    Parameters
    ----------
    info : esm_tests.Info
        Dictionary that contains the testing info.
    """
    # Load dictionaries from ``info``
    scripts_info = info["scripts"]
    user_info = info["user"]
    this_computer = info["this_computer"]

    # Set the regex format for check compilations
    cd_format = re.compile("         cd (.*)")

    # Set the counter to 0
    c = 0
    logger.info("")
    # Loop through the models and their scripts
    for model, scripts in scripts_info.items():
        if model == "general":
            continue
        # Loop through the scripts of each model
        for script, v in scripts.items():
            # If the user has specified the ``hold`` flag, wait for input
            if info["hold"] and c != 0:
                input("Press ENTER to continue...")
            # Calculate progress
            c += 1
            progress = round(c / scripts_info["general"]["num_scripts"] * 100, 1)

            # Define variables from the runscript
            version = v["version"]
            if v["comp_command"]:
                comp_command = f"{v['comp_command']} --no-motd -k"
            else:
                comp_command = f"esm_master comp-{model}-{version} --no-motd -k"
            general_model_dir = f"{user_info['test_dir']}/comp/{model}"
            model_dir = f"{user_info['test_dir']}/comp/{model}/{model}-{version}"
            title = f"COMPILING ({progress}%) {bs}{model}-{version}{be}"
            logger.info("\t" + info["group_output"]["startg"].format(title))

            # Create the folders where this test should run and change directory
            if not os.path.isdir(general_model_dir):
                os.makedirs(general_model_dir)
            os.chdir(general_model_dir)

            # Check if the test compilation has been already run (the compilation
            # directory already exists). This is particularly important for different
            # testing scripts that use the same version of a model, so that compilation
            # is not run more than once
            if os.path.isdir(model_dir):
                v["action"] = {"comp": "Directory already exists"}
                logger.info(f"\t\tDirectory already exists, skipping")
                with open(f"{model_dir}/comp.out") as o:
                    out = o.read()
            else:
                # ACTUAL COMPILATION
                # ==================
                # Get the source code if actual compilation is required
                if info["actually_compile"]:
                    # Downloading or copying
                    # ----------------------
                    # Load the ``config.yaml`` for this model to find if the source
                    # code needs to be downloaded or copied
                    with open(f"{os.path.dirname(v['path'])}/config.yaml", "r") as cf:
                        config_test = yaml.load(cf, Loader=yaml.FullLoader)
                    copying = (
                        config_test["comp"]
                        .get("cp_instead_of_download", {})
                        .get(this_computer, {})
                        .get(f"{model}-{version}")
                    )
                    # If copying of the source code is required go ahead, copy and,
                    # if necessary, untar
                    if copying:
                        logger.info("\t\tCopying")
                        out = ""
                        if os.path.isdir(copying):
                            shutil.copytree(copying, general_model_dir)
                        else:
                            shutil.copy2(copying, general_model_dir)
                        if copying.endswith("tar.gz"):
                            logger.info("\t\tUntaring")
                            file2untar = (
                                f"{general_model_dir}/{os.path.basename(copying)}"
                            )
                            out += sh(f"tar -xvf {file2untar}")
                            os.remove(file2untar)
                    # If download of the source code is required, run
                    # ``esm_master get-<model>-<version>``
                    else:
                        get_command = f"esm_master get-{model}-{version} --no-motd"
                        logger.info("\t\tDownloading")
                        out = sh(get_command)
                        if "Traceback (most recent call last):" in out:
                            logger.error(f"\t\t\tProblem downloading!\n\n{out}")

                # CHECK COMPILATION
                # =================
                # No actual compilation: for a check compilation where no real
                # compilation occurs, but the ``comp-`` scripts are generated, we need
                # to trick esm_master into thinking that all the folders source code
                # folders are available, even if they were not downloaded or copied
                else:
                    # Evaluate and create folders to trick esm_master
                    # -----------------------------------------------
                    # Run a ``esm_master --check``
                    out = sh(f"{comp_command} -c")
                    folders = []
                    # Search for the folders to be created
                    for line in out.split("\n"):
                        if "cd" in line and "cd .." not in line:
                            found_format = cd_format.findall(line)
                            if len(found_format) > 0:
                                if (
                                    ";" not in found_format[0]
                                    and "/" not in found_format[0]
                                ):
                                    folders.append(found_format[0])
                    if len(folders) == 0:
                        logger.warning(
                            f'NOT TESTING {model + version}: "cd" command not found'
                        )
                        continue

                    # Make the folders needed to trick esm_master
                    prim_f = folders[0]
                    folders.append(f"{model}-{version}")
                    folders = [x for x in set(folders)]
                    os.mkdir(prim_f)
                    for folder in folders:
                        os.mkdir(prim_f + "/" + folder)

                    # Get files from the
                    # ``esm_test/runscripts/<model>/comp_files4check_runs`` (i.e.
                    # namelists that are hosted in another repository could be placed
                    # there so that checks can run successfully without having to
                    # download the code)
                    copy_comp_files4check_runs(
                        script, v, f"{general_model_dir}/{prim_f}"
                    )

                # Compile
                if info["actually_compile"]:
                    logger.info("\t\tCompiling")
                else:
                    logger.info("\t\tWritting compilation scripts")
                out = sh(comp_command)

                # Write output file
                with open(f"{model_dir}/comp.out", "w") as o:
                    o.write(out)

                # Move and cleanup files
                for f in os.listdir(general_model_dir):
                    if "comp-" in f and not os.path.isfile(f"{model_dir}/{f}"):
                        shutil.move(f"{general_model_dir}/{f}", model_dir)
                    elif "comp-" in f and os.path.isfile(f"{model_dir}/{f}"):
                        os.remove(f"{general_model_dir}/{f}")
                    if f == "dummy_script.sh":
                        os.remove(f"{general_model_dir}/{f}")

            # Checks
            success = check(info, "comp", model, version, out, script, v)
            if success:
                logger.info("\t\tSuccess!")
            logger.info(f'\t\t{info["group_output"]["endg"]}')


def run_test(info):
    """
    Runs the simulations and notifies when they are finished. If the ``--check`` flag
    is given in the ``esm_tests`` call, check simulations are run
    (``esm_runscripts -c``).

    Parameters
    ----------
    info : esm_tests.Info
        Dictionary that contains the testing info.
    """
    # Load dictionaries from ``info``
    scripts_info = info["scripts"]
    user_info = info["user"]
    actually_run = info["actually_run"]
    run_errors = ["ERROR:", "slurmstepd: error: *** STEP"]

    # Set the counter to 0
    c = 0
    submitted = []
    # Loop through the models and their scripts
    for model, scripts in scripts_info.items():
        if model == "general":
            continue
        # Loop through the scripts of each model
        for script, v in scripts.items():
            # If the user has specified the ``hold`` flag, wait for input
            if info["hold"] and c != 0:
                input("Press ENTER to continue...")
            c += 1
            progress = round(c / scripts_info["general"]["num_scripts"] * 100, 1)

            # Define variables from the runscript
            version = v["version"]
            runscript_path = v["path"]
            general_run_dir = f"{user_info['test_dir']}/run/{model}/"
            run_dir = f"{general_run_dir}/{script}"
            model_dir = f"{user_info['test_dir']}/comp/{model}/{model}-{version}"
            title = f"SUBMITTING ({progress}%) {bs}{model}/{script}{be}"
            logger.info(f"\t" + info["group_output"]["startg"].format(title))

            # Create the folders where this test should run and change directory
            if not os.path.isdir(general_run_dir):
                os.makedirs(general_run_dir)

            # If the simulation already exists, only read the ``.out`` file
            if os.path.isdir(run_dir):
                v["action"]["submission"] = "Directory already exists"
                logger.info(f"\t\tDirectory already exists, skipping")
                with open(f"{run_dir}/run.out", "r") as o:
                    out = o.read()
                if actually_run:
                    submitted.append((model, script))
            # If the simulation does not exists, run it or check it
            else:
                # Change the path to that of the runscript
                os.chdir(os.path.dirname(runscript_path))

                # Define the check flag if necessary
                additional_actions = ""
                if actually_run:
                    check_flag = ""
                else:
                    check_flag = "-c"
                    # In case of iterative coupling, submit the subscripts too
                    if v["iterative_coupling"]:
                        for imodel, mdata in v["iterative_models"].items():
                            additional_actions = (
                                f"{additional_actions}; "
                                f"echo {'='*10}; echo {imodel}; echo {'='*10}; "
                                f"esm_runscripts "
                                f"{os.path.dirname(v['path'])}/{mdata['script']} "
                                f"-e {script} --open-run {check_flag}"
                            )
                # Export test variables and run the simulation
                env_vars = [
                    f"ACCOUNT='{user_info['account']}'",
                    f"ESM_TESTING_DIR='{general_run_dir}'",
                    f"MODEL_DIR='{model_dir}'",
                ]
                run_command = (
                    f"esm_runscripts {v['path']} -e {script} --open-run "
                    f"{check_flag}{additional_actions}"
                )
                out = sh(run_command, env_vars)

                # Write output file
                if os.path.isdir(run_dir):
                    with open(f"{run_dir}/run.out", "w") as o:
                        o.write(out)

                if actually_run:
                    submitted.append((model, script))

            # Check submission
            success = check(info, "submission", model, version, out, script, v)

            logger.info(f'\t\t{info["group_output"]["endg"]}')

    # Check if simulations are finished
    total_sub = len(submitted)
    subc = 1
    if total_sub > 0:
        logger.info(
            "\nWaiting for submitted runs to finish... You can choose to cancel the "
            + "script now and come back to it at a later state with the same command "
            + "you submitted it, but remember to remove the '-d' so that nothing is "
            + "deleted and this script can be resumed from where it was."
        )
    infoc = 10
    while submitted:
        cc = 0
        finished_runs = []
        # Loop through models and scripts
        for model, script in submitted:
            v = scripts_info[model][script]
            version = v["version"]
            progress = round(subc / total_sub * 100, 1)
            exp_dir = f"{user_info['test_dir']}/run/{model}/{script}/"
            exp_dir_log = f"{exp_dir}/log/"
            nmodels_success = 1
            with open(f"{exp_dir}/run.out", "r") as so:
                submission_out = so.read()
            if "ERROR:" in submission_out or not v["state"].get("submission", False):
                subc, finished_runs, success = experiment_state_action(
                    info,
                    "Submission failed!",
                    False,
                    v,
                    finished_runs,
                    cc,
                    subc,
                    model,
                    script,
                    version,
                    progress,
                )
            # Search through the ``_compute_`` files for a string that indicates that
            # the run has finished
            for f in os.listdir(exp_dir_log):
                if "_compute_" in f and ".log" in f:
                    with open(f"{exp_dir_log}/{f}") as m:
                        observe_out = m.read()
                        # If the run has finished successfully label the state for
                        # ``run_finished`` as ``True`` and run a check for files that
                        # should have been created
                        if "Reached the end of the simulation, quitting" in observe_out:
                            if nmodels_success == v["nmodels_iterative_coupling"]:
                                subc, finished_runs, success = experiment_state_action(
                                    info,
                                    "Success!",
                                    True,
                                    v,
                                    finished_runs,
                                    cc,
                                    subc,
                                    model,
                                    script,
                                    version,
                                    progress,
                                )
                            else:
                                nmodels_success += 1
                        # If the run has errors label the state for ``run_finished`` as
                        # ``False`` and run a check for files that should have been
                        # created anyway
                        elif any([run_error in observe_out for run_error in run_errors]):
                            subc, finished_runs, success = experiment_state_action(
                                info,
                                "Simulation crashed!",
                                False,
                                v,
                                finished_runs,
                                cc,
                                subc,
                                model,
                                script,
                                version,
                                progress,
                            )

            # Update the testing index for the next iteration
            cc += 1

        # Remove finished runs from the ``submitted`` list
        finished_runs = list(set(finished_runs))
        for indx in finished_runs[::-1]:
            del submitted[indx]

        # Wait 10 iterations to notify the state
        if infoc == 10 and len(submitted) > 0:
            infoc = 0
            runs = ""
            for model, script in submitted:
                runs += f"\t- {model}/{script}\n"
            logger.info(f"\nWaiting for the following runs to finish:\n{runs}")
        else:
            infoc += 1

        # Wait 30 seconds to check again the state of the submitted tests
        if len(submitted) > 0:
            time.sleep(30)

    logger.info("")


#######################################################################################
# CHECKS
#######################################################################################
def check(info, mode, model, version, out, script, v):
    """
    Checks for errors, missing files and compares current files with previous files in
    ``resources/last_tested``. Behaves differently for different modes.

    Modes:
        * ``comp``: checks compilation scripts
        * ``submission``: checks submission files (``.run``, namelists, ...)
        * ``run``: same as ``submission`` but also checks for files produced by the
            simulation (restarts and output)

    Parameters
    ----------
    info : esm_tests.Info
        Dictionary that contains the testing info.
    mode : str
        Mode of the check.
    model : str
        Model for which the check is run.
    version : str
        Version of the model for which the check is run.
    out : str
        Output of the shell operation that is to be checked.
    script : str
        Name of the script to be checked.
    v : dict
        Dictionary containing script variables such as ``version``, ``comp_command``, ...

    Returns
    -------
    success : bool
        ``True`` if all checks were passed, ``False`` if not.
    """
    # Set variables
    success = True
    mode_name = {"comp": "compilation", "submission": "submission", "run": "runtime"}
    last_tested_dir = info["last_tested_dir"]
    this_computer = info["this_computer"]
    user_info = info["user"]
    actually_run = info["actually_run"]

    # Load config for this mode
    with open(f"{os.path.dirname(v['path'])}/config.yaml", "r") as c:
        config_test = yaml.load(c, Loader=yaml.FullLoader)

    # Build the ignore_dict based on the general, model and script specific ignores
    ignore_dict = copy.deepcopy(info["ignore"])
    model_ignore = config_test.get("ignore_compare", {}).get("all", {})
    deep_update(ignore_dict, model_ignore, extend_lists=True)
    script_ignore = config_test.get("ignore_compare", {}).get(script, {})
    deep_update(ignore_dict, script_ignore, extend_lists=True)

    # There are 3 modes: comp, submission and run, and 2 config_modes: comp and run.
    # This is because submission shares a lot of operations with run so they are in the
    # same config_mode group
    if mode == "submission":
        config_mode = "run"
    else:
        config_mode = mode

    # Check if the config_mode is defined in the ``config.yaml`` file, and load the
    # info for this type of test
    if config_mode not in config_test:
        logger.error(
            f"Missing '{mode}' section in '{os.path.dirname(v['path'])}/config.yaml'!"
        )
    config_test = config_test[config_mode]

    # Set mode variables
    if mode == "comp":
        if info["actually_compile"]:
            test_type = "actual"
        else:
            test_type = "check"
        actually_do = info["actually_compile"]
        subfolder = f"{model}-{version}"
    elif mode == "run":
        if actually_run:
            test_type = "actual"
        else:
            test_type = "check"
        actually_do = actually_run
        subfolder = script
    elif mode == "submission":
        # Do not perform the file checks before the simulation is finished
        if actually_run:
            test_type = "actual"
        else:
            test_type = "check"
        actually_do = False
        subfolder = script

    # Check for errors
    if actually_do or mode == "submission":
        # Check for errors in the output
        errors = config_test.get(test_type, {}).get("errors", [])
        # Get exceptions
        error_exceptions = {}
        cleaned_errors = []
        for error in errors:
            cleaned_error = error.split(" except ")[0]
            error_exceptions[cleaned_error] = find_exceptions(error)
            cleaned_errors.append(cleaned_error)
        errors = cleaned_errors
        # Add specific errors
        if mode == "comp":
            errors.append("errors occurred!")
        if mode == "submission":
            errors.extend(["Traceback (most recent call last):", "ERROR"])
        # Loop through errors
        for error in errors:
            if error in out and script not in error_exceptions.get(error, []):
                logger.error(f"\t\tError during {mode_name[mode]}!\n\n{out}")
                success = False
        if mode != "run":
            v["state"][mode] = success
    else:
        if mode == "comp":
            # No need to check for this in check mode
            v["state"][mode] = success

    # Check for missing files during an actual operation (not a check, not submission)
    if actually_do:
        # Check if files exist
        files_checked = exist_files(
            config_test.get(test_type, {}).get("files", []),
            f"{user_info['test_dir']}/{mode}/{model}/{subfolder}",
            version,
        )
        v["state"][f"{mode}_files"] = files_checked
        success = success and files_checked

    # Compare scripts with previous, if existing
    this_compare_files = copy.deepcopy(compare_files[config_mode])
    # TODO: The iterative coupling needs a rework. Therefore, no testing for files
    # is develop. Include the tests after iterative coupling is reworked
    if config_mode == "run" and v["iterative_coupling"]:
        this_compare_files = []
    this_compare_files.extend(config_test.get(test_type, {}).get("compare", []))
    this_test_dir = f"{config_mode}/{model}/{subfolder}/"
    v["state"][f"{mode}_files_identical"] = True
    # Loop through the files to be compared
    for cfile in this_compare_files:
        # Load lines to be ignored
        ignore_lines = ignore_dict.get(cfile, [])
        # Get relative paths of the files to be compared
        subpaths_source, subpaths_target = get_rel_paths_compare_files(
            info, cfile, v, this_test_dir
        )
        for sp, sp_t in zip(subpaths_source, subpaths_target):
            # Check if the source exists (this simulation file)
            if not os.path.isfile(f"{user_info['test_dir']}/{sp}"):
                logger.error(f"\t\t'{sp}' file is missing!")
                identical = False
            else:
                # Check if the target exists (in last_tested folder)
                if os.path.isfile(f"{last_tested_dir}/{this_computer}/{sp_t}"):
                    # Check if files are identical
                    identical, differences = print_diff(
                        info,
                        f"{last_tested_dir}/{this_computer}/{sp_t}",
                        f"{user_info['test_dir']}/{sp}",
                        sp,
                        ignore_lines,
                    )
                    success += identical
                    # Update state dictionaries
                    if not identical:
                        v["differences"] = v.get("differences", {})
                        v["differences"][config_mode] = v["differences"].get(
                            config_mode, {}
                        )
                        v["differences"][config_mode][sp] = differences
                        v["state"][f"{mode}_files_identical"] &= False
                    else:
                        v["state"][f"{mode}_files_identical"] &= True
                else:
                    logger.warning(f"\t\t'{sp_t}' file not yet in 'last_tested'")

    return success


def experiment_state_action(
    info, message, no_err, v, finished_runs, cc, subc, model, script, version, progress
):
    """
    Triggers checks for finished runs and reports their state.

    Parameters
    ----------
    info : esm_tests.Info
        Dictionary that contains the testing info
    message : str
        Message to report about the state
    no_err : bool
        Boolean indicating whether errors occur (``False``) or not (``True``)
    v : dict
        Dictionary containing script variables such as ``version``, ``comp_command``, ...
    finished_runs : list
        List of indexes for runs finished
    cc : int
        Run counter
    subc : int
        Counter for subruns
    model : str
        Model for which the check is run
    script : str
        Name of the script to be checked
    version : str
        Version of the model for which the check is run
    progress : float
        Percentage of the progress

    Returns
    -------
    subc : int
        Counter for subruns
    finished_runs : list
        List of indexes for runs finished
    success : bool
        Boolean indicating the success of the run
    """
    logger.info(f"\tRUN FINISHED ({progress}%) {model}/{script}")
    if no_err:
        logger.info(f"\t\t{message}")
    else:
        logger.error(f"\t\t{message}")
    finished_runs.append(cc)
    subc += 1
    v["state"]["run_finished"] = no_err
    # Run a check for run finished
    success = check(info, "run", model, version, "", script, v)

    return subc, finished_runs, success


def check_perfect(info, results):
    """
    Exits if something is not perfect in the ``results``, to trigger a failted test
    in GitHub Actions.

    Parameters
    ----------
    info : esm_tests.Info
        Dictionary with the general info about the tests.
    results : dict
        Dictionary containing the results of the testing.
    """
    all_tests_passed = True
    if info["actually_compile"]:
        comp_perfect = "compiles"
    else:
        comp_perfect = "comp files identical"
    if info["actually_run"]:
        run_perfect = "runs"
    else:
        run_perfect = "run files identical"
    for model, versions in results.items():
        for version, scripts in versions.items():
            for script, computers in scripts.items():
                for computer, data in computers.items():
                    if data["compilation"] != comp_perfect:
                        all_tests_passed = False
                    if data["run"] != run_perfect:
                        all_tests_passed = False
    if not all_tests_passed:
        sys.exit(
            "Some of the tests were not successful. Exited to trigger a failed test in GitHub Actions"
        )


def exist_files(files, path, version):
    """
    Checks whether the required files defined in
    ``resources/runscripts/<model>/config.yaml`` are present or not and sets the state
    if the ``files_check`` ``True`` (exist) or ``False`` (do not exist) depending on
    that.

    Each element of the list of files in the ``config.yaml`` can contain the following
    special functionalities:
    - ``*``: to trigger a wildcard evaluation
    - ``except [<list of versions>]``: do not check this file for the versions inside
      the ``[]``
    - ``in [<list of versions>]``: check this file only for the versions inside the
      ``[]``

    Parameters
    ----------
    files : list
        List of files to be checked, extracted from the ``config.yaml``
    path : str
        Path of the experiment
    version : str
        Version of the model
    """
    files_checked = True
    # Loop through files
    for f in files:
        exception_list = find_exceptions(f)

        # Command's logic
        if " except " in f and version in exception_list:
            continue
        elif " in " in f and not version in exception_list:
            continue
        else:
            f_path = f.split(" ")[0]

        # Check for files with wildcards
        if "*" in f_path:
            listing = glob.glob(f"{path}/{f_path}")
            if len(listing) == 0:
                logger.error(
                    f"\t\tNo files following the pattern '{f_path}' were created!"
                )
                files_checked = False
        # Check for files without wildcards
        else:
            if not os.path.isfile(f"{path}/{f_path}"):
                logger.error(f"\t\t'{f_path}' does not exist!")
                files_checked = False

    return files_checked

def find_exceptions(string_to_be_checked):
    """
    Finds exceptions on the ``string_to_be_checked`` with the format
    `` except [<item1>, <item2>, ...]``, and returns the list of the exceptions
    specified between the square brackets in the string.

    Parameters
    ----------
    string_to_be_checked : str
        String to be searched for exceptions

    Returns
    -------
    exception_list : list
        List of exceptions found on the string
    """
    exception_list = []
    # Get the commands inside the ``[]``
    if " [" in string_to_be_checked and string_to_be_checked[-1] == "]":
        exception_list = re.findall(r"(?<=\[)([^]]+)(?=\])", string_to_be_checked)
        if len(exception_list) > 1:
            raise Exception("You should only have one list per file")
        exception_list = [x.replace(" ", "") for x in exception_list[0].split(",")]

    return exception_list



import os
import re
import sys

import yaml
from loguru import logger

import esm_parser

YAML_AUTO_EXTENSIONS = ["", ".yml", ".yaml", ".YML", ".YAML"]


class EsmConfigFileError(Exception):
    """
    Exception for yaml file containing tabs or other syntax issues.

    An exception used when yaml.load() throws a yaml.scanner.ScannerError.
    This error occurs mainly when there are tabs inside a yaml file or
    when the syntax is incorrect. If tabs are found, this exception returns
    a user-friendly message indicating where the tabs are located in the
    yaml file.

    Parameters
    ----------
    fpath : str
        Path to the yaml file
    """

    def __init__(self, fpath, yaml_error):
        report = ""
        # Loop through the lines inside the yaml file searching for tabs
        with open(fpath) as yaml_file:
            for n, line in enumerate(yaml_file):
                # Save lines and line numbers with tabs
                if "\t" in line:
                    report += str(n) + ":" + line.replace("\t", "____") + "\n"

        # Message to return
        if len(report) == 0:
            # If no tabs are found print the original error message
            print("\n\n\n" + yaml_error)
        else:
            # If tabs are found print the report
            self.message = (
                "\n\n\n"
                f"Your file {fpath} has tabs, please use ONLY spaces!\n"
                "Tabs are in following lines:\n" + report
            )
        super().__init__(self.message)


# This next part is stolen here:
# https://medium.com/swlh/python-yaml-configuration-with-environment-variables-parsing-77930f4273ac
# Deniz: unfortunately this example as well as the other examples and even
# PyYaml itself is bit buggy. Basically everything with ${VAR} passes through
# the pipeline and this ends up replacing anything with ${VAR} with VAR and
# export them as environmental variables. That propagates even further and
# messes up the rest of the esm_tools (eg. master, runscripts, ...). The fix
# below corrects that issue.
def create_env_loader(tag="!ENV", loader=yaml.SafeLoader):
    # pattern for ${VAR} and extract VAR
    pattern_envvar = re.compile("\${(\w+)}")

    # This pattern matches the valid (uncommented) !ENV lines. Eg.
    # - !ENV ${VAR}
    # - !ENV "${VAR}" or !ENV '${VAR}'
    # Note: '!ENV ...' or "!ENV ..." does not work since PyYaml requires the tag outside of the quotes
    pattern_envtag = re.compile("""^[^\#]*\!ENV[ \t]+['|"]?\$\{\w+\}['|"]?""")

    # the tag will be used to mark where to start searching for the pattern
    # e.g. somekey: !ENV somestring${MYENVVAR}blah blah blah
    loader.add_implicit_resolver(tag, pattern_envvar, None)
    loader.env_variables = []

    def constructor_env_variables(loader, node):
        """
        Extracts the environment variable from the node's value
        :param yaml.Loader loader: the yaml loader
        :param node: the current node in the yaml
        :return: the parsed string that contains the value of the environment
        variable
        """
        # file name (including the full path if not in PWD)
        fname = node.start_mark.name

        # line number (starts from 0) where match is found
        line_num = node.start_mark.line

        # start and end of the colums numbers in the line (starts from 0)
        # For character index from the start of the file use node.start_mark.index
        column_start = node.start_mark.column
        column_end = node.end_mark.column

        # read the file into a list of line. This can also be achieved by
        # read() instead but since the YAML files are not large (max few KBs),
        # this is a quicked solution
        yaml_file = open(fname, "r")
        file_lines = yaml_file.readlines()
        yaml_file.close()

        # current line without the newline at the end
        cur_line = file_lines[line_num].rstrip()

        # Print extra debug messages if ESM_PARSER_DEBUG environment variable is set
        if os.getenv("ESM_PARSER_DEBUG"):
            print()
            print("===== esm_parser -> yaml_to_dict.py -> create_env_loader() =====")
            print(f"::: Reading the file: {fname}")
            print("::: Reading the line:")
            print(f">>>{cur_line}<<<")
            print(f"::: found the match: >>>{cur_line[column_start:column_end]}<<<")
            print("==================== END OF ESM_PARSER_DEBUG ===================")
            print()

        value = loader.construct_scalar(node)

        # Check if we have a valid !ENV tag on the line
        envtag_match = re.search(pattern_envtag, cur_line)

        # list of matches for ${VAR}. Each match is VAR
        envvar_matches = pattern_envvar.findall(value)

        # Parse the environmental variables only if the line has a valid !ENV tag
        if envtag_match:
            if envvar_matches:
                full_value = value
                for env_var in envvar_matches:

                    # first check if the variable exists in the shell environment
                    if not os.getenv(env_var):
                        esm_parser.user_error(
                            f"{env_var} is not defined",
                            f"{env_var} is not an environment variable. Exiting",
                        )

                    # replace {env_var} with the value of the env_var
                    full_value = full_value.replace(
                        f"${{{env_var}}}", os.getenv(env_var)
                    )
                    loader.env_variables.append((env_var, full_value))
                return full_value
        return value

    loader.add_constructor(tag, constructor_env_variables)
    return loader


def yaml_file_to_dict(filepath):
    """
    Given a yaml file, returns a corresponding dictionary.

    If you do not give an extension, tries again after appending one.
    It raises an EsmConfigFileError exception if yaml files contain tabs.

    Parameters
    ----------
    filepath : str
        Where to get the YAML file from

    Returns
    -------
    dict
        A dictionary representation of the yaml file.

    Raises
    ------
    EsmConfigFileError
        Raised when YAML file contains tabs or other syntax issues.
    FileNotFoundError
        Raised when the YAML file cannot be found and all extensions have been tried.
    """
    loader = create_env_loader()
    for extension in YAML_AUTO_EXTENSIONS:
        try:
            with open(filepath + extension) as yaml_file:
                # Check for duplicates
                check_duplicates(yaml_file)
                # Back to the beginning of the file
                yaml_file.seek(0, 0)
                # Actually load the file
                yaml_load = yaml.load(yaml_file, Loader=loader)  # yaml.FullLoader)
                # Check for incompatible ``_changes`` (no more than one ``_changes``
                # type should be accessible simultaneously)
                check_changes_duplicates(yaml_load, filepath + extension)
                # Add the file name you loaded from to track it back:
                yaml_load["debug_info"] = {"loaded_from_file": yaml_file.name}
                if loader.env_variables:
                    runtime_env_changes = yaml_load.get("computer", {}).get(
                        "runtime_environment_changes", {}
                    )
                    add_export_vars = runtime_env_changes.get("add_export_vars", {})
                    for env_var_name, env_var_value in loader.env_variables:
                        add_export_vars[env_var_name] = env_var_value
                    # TODO(PG): There is probably a more elegant way of doing this:
                    yaml_load["computer"] = yaml_load.get("computer") or {}
                    yaml_load["computer"]["runtime_environment_changes"] = (
                        yaml_load["computer"].get("runtime_environment_changes") or {}
                    )
                    yaml_load["computer"]["runtime_environment_changes"][
                        "add_export_vars"
                    ] = add_export_vars
                # Check for empty components/models
                # check_for_empty_components(yaml_load, filepath + extension)
                return yaml_load
        except IOError as error:
            logger.debug(
                f"IOError ({error.errno}): File not found with {filepath+extension}, trying another extension pattern."
            )
        except yaml.scanner.ScannerError as yaml_error:
            logger.debug(f"Your file {filepath + extension} has syntax issues!")
            raise EsmConfigFileError(filepath + extension, yaml_error)
        except Exception as error:
            logger.exception(error)
            esm_parser.user_error(
                "Yaml syntax",
                f"Syntax error in ``{filepath}``\n\n``Details:\n``{error}")
    raise FileNotFoundError(
        "All file extensions tried and none worked for %s" % filepath
    )


def check_changes_duplicates(yamldict_all, fpath):
    """
    Checks for duplicates and conflicting ``_changes`` and ``add_``:

    1. Finds variables containing ``_changes`` (but excluding ``add_``) and checks
       if they are compatible with the same ``_changes`` inside the same file. If they
       are not compatible returns an error where the conflicting variable paths are
       specified. More than one ``_changes`` type in a file are allowed but they need
       to be part of the same ``_choose`` and not be accessible simultaneously in any
       situation.

    2. Checks if there is any variable containing ``add_`` in the main sections of
       a file and labels it as incompatible if the same variable is found inside a
       ``choose_`` block. ``add_<variable>``s are compatible as long as they are inside
       ``choose_`` blocks, but if you want to include something as a default, please just
       do it inside the ``<variable>``.

       .. warning:: ``add_<variable>``s are not checked for incompatibility when they
          are included inside ``choose_`` blocks. Merging of these ``add_<variable>``s
          is done using ``deep_update``, meaning that the merge is arbitrary (i.e. if
          two ``choose_`` blocks are modifying the same variable using ``add_``, the
          final value would be decided arbitrarily). It is up to the developer/user to
          make good use of ``add_``s inside ``choose_`` blocks.

    Parameters
    ----------
    yamldict_all : dict
        Dictionary read from the yaml file
    fpath : str
        Path to the yaml file
    """
    changes_note = (
        "Note that if there are more than one ``_changes`` in the "
        "file, they need to be placed inside different cases of the "
        "same ``choose`` and these options need to be compatible "
        "(only one ``_changes`` can be reached at a time).\n"
        "Use ``add_<variable>_changes`` if you want to add/overwrite "
        "variables inside the main ``_changes``."
    )
    add_note = (
        "Note that multiple ``add_<variable>`` in a single file are compatible "
        "as long as they are included inside ``choose_`` blocks. An "
        "``add_<variable>`` out of a ``choose_`` block and the same "
        "``add_<variable>`` inside of a ``choose_`` block are considered "
        "incompatible. If the general ``add_<variable>`` should be added "
        "as a default, please include it to ``<variable>`` instead."
    )

    # If it is a couple setup, check for ``_changes`` duplicates separately for each component
    if "general" not in yamldict_all:
        yamldict_all = {"main": yamldict_all}

    # Loop through the components or main
    for yamldict in yamldict_all.values():
        # Check if any <variable>_changes or add_<variable> exists, if not, return
        # Perform the check only for the dictionary objects
        changes_list = []
        add_list = []
        if isinstance(yamldict, dict):
            changes_list = esm_parser.find_key(
                yamldict, "_changes", "add_", paths2finds=[], sep=","
            )
            add_list = esm_parser.find_key(
                yamldict, ["add_"], "", paths2finds=[], sep=","
            )
            if (len(changes_list) + len(add_list)) == 0:
                continue

        # Find ``_changes`` types
        changes_types = set(
            [y for x in changes_list for y in x.split(",") if "_changes" in y]
        )
        # Find ``add_`` types
        add_types = set([y for x in add_list for y in x.split(",") if "add_" in y])
        # Define ``_changes`` groups
        changes_groups = []
        for change_type in changes_types:
            changes_groups.append(
                [x for x in changes_list if change_type == x.split(",")[-1]]
            )
        # Define ``add_`` groups
        add_groups = []
        for add_type in add_types:
            add_groups.append([x for x in add_list if add_type == x.split(",")[-1]])

        # Loop through the different ``_changes`` groups
        for changes_group in changes_groups:
            # Check for ``_changes`` without ``choose_``, "there can be only one"
            changes_no_choose = [x for x in changes_group if "choose_" not in x]
            # If more than one ``_changes`` without ``choose_`` return error
            if len(changes_no_choose) > 1:
                changes_no_choose = [x.replace(",", ".") for x in changes_no_choose]
                esm_parser.user_error(
                    "YAML syntax",
                    "More than one ``_changes`` out of a ``choose_``in "
                    + fpath
                    + ":\n    - "
                    + "\n    - ".join(changes_no_choose)
                    + "\n"
                    + changes_note
                    + "\n\n",
                )
            # If only one ``_changes`` without ``choose_`` check for ``_changes`` inside
            # ``choose_`` and return error if any is found
            elif len(changes_no_choose) == 1:
                changes_group.remove(changes_no_choose[0])
                if len(changes_group) > 0:
                    changes_group = [x.replace(",", ".") for x in changes_group]
                    esm_parser.user_error(
                        "YAML syntax",
                        "The general ``"
                        + changes_no_choose[0]
                        + "`` and ``_changes`` in ``choose_`` are not compatible in "
                        + fpath
                        + ":\n    - "
                        + "\n    - ".join(changes_group)
                        + "\n"
                        + "\n"
                        + changes_note
                        + "\n\n",
                    )

            # If you reach this point all ``_changes`` are inside
            # some number of ``choose_`` (there are no ``_changes``
            # outside of a ``choose_``)

            # Check for incompatible ``_changes`` inside ``choose_``:
            # Split the path of the variables
            changes_group_split = [x.split(",") for x in changes_group]
            # Loop through the paths of the ``_changes`` in the group
            for count, changes in enumerate(changes_group_split):
                # Find the path of the last ``choose_`` in ``changes`` and
                # its case
                path2choose, case = find_last_choose(changes)
                # Loop through the changes following the current one
                for other_changes in changes_group_split[count + 1 :]:
                    # Find the path of the last ``choose_`` in
                    # ``other_changes`` and its case
                    sub_path2choose, sub_case = find_last_choose(other_changes)
                    # If one ``choose_`` is contained into the other
                    # find the common ``choose_`` and compare the cases.
                    # If the case is the same, duplicates exist and error
                    # is returned (i.e. choose_lresume.True.namelist_changes
                    # and choose_lresume.True.choose_another_switch
                    # False.namelist_changes)
                    if path2choose in sub_path2choose or sub_path2choose in path2choose:
                        if path2choose in sub_path2choose:
                            sub_case = sub_path2choose.replace(
                                path2choose + ",", ""
                            ).split(",")[0]
                        elif sub_path2choose in path2choose:
                            case = path2choose.replace(sub_path2choose + ",", "").split(
                                ","
                            )[0]
                        if case == sub_case:
                            esm_parser.user_error(
                                "YAML syntax",
                                "The following ``_changes`` can be accessed "
                                + "simultaneously in "
                                + fpath
                                + ":\n"
                                + "    - "
                                + ".".join(changes)
                                + "\n"
                                + "    - "
                                + ".".join(other_changes)
                                + "\n"
                                + "\n"
                                + changes_note
                                + "\n\n",
                            )
                    else:
                        # If these ``choose_`` are different they can be accessed
                        # simultaneously, then it returns an error
                        esm_parser.user_error(
                            "YAML syntax",
                            "The following ``_changes`` can be accessed "
                            + "simultaneously in "
                            + fpath
                            + ":\n"
                            + "    - "
                            + ".".join(changes)
                            + "\n"
                            + "    - "
                            + ".".join(other_changes)
                            + "\n"
                            + "\n"
                            + changes_note
                            + "\n\n",
                        )

        # Loop through the different ``add_`` groups
        for add_group in add_groups:
            # Count ``add_`` occurrences out of a ``choose_``
            add_no_choose = [x for x in add_group if "choose_" not in x]
            # Check if the heading of the path is the same. If not is not a
            # repetition
            for anc in add_no_choose:
                add_no_choose = []
                heading = anc.split(",")[0]
                counter = 0
                for ag in add_group:
                    if heading in ag:
                        counter += 1
                if counter > 1:
                    add_no_choose.append(anc)

            # If one ``add_`` without ``choose_`` check for ``add_`` inside
            # ``choose_`` and return error if any is found (incompatible ``add_``s)
            if len(add_no_choose) >= 1:
                add_group.remove(add_no_choose[0])
                if len(add_group) > 0:
                    add_group = [x.replace(",", ".") for x in add_group]
                    esm_parser.user_error(
                        "YAML syntax",
                        "The general ``"
                        + add_no_choose[0]
                        + "`` and ``add_`` in ``choose_`` are not compatible in "
                        + fpath
                        + ":\n    - "
                        + "\n    - ".join(add_group)
                        + "\n\n"
                        + add_note
                        + "\n\n",
                    )


def check_for_empty_components(yaml_load, fpath):
    for key, value in yaml_load.items():
        if not value:
            esm_parser.user_error(
                "YAML syntax",
                f"The component ``{key}`` is empty in the file ``{fpath}``. ESM-Tools does"
                + " not support empty components, either add some variables to the "
                + f"``{key}`` section, or remove it from this file.",
            )


def find_last_choose(var_path):
    """
    Locates the last ``choose_`` on a string containing the path to a
    variable separated by ",", and returns the path to the ``choose_``
    (also separated by ",") and the case that follows the ``choose_``.

    Parameters
    ----------
    var_path : str
        String containing the path to the last ``choose_`` separated by
        ",".

    Returns
    -------
    path2choose : str
        Path to the last ``choose_``.
    case : str
        Case after the choose.
    """
    # Find the last ``choose_``
    last_choose = [x for x in var_path if "choose_" in x][-1]
    # Find the last ``choose_`` index
    choose_index = var_path.index(last_choose)
    # Defines the path to the last ``choose_``
    path2choose = ",".join(var_path[: var_path.index(last_choose) + 1])
    # Defines the case of the last ``choose_``
    case = var_path[choose_index + 1]
    return path2choose, case


def check_duplicates(src):
    """
    Checks that there are no duplicates in a yaml file, and if there are
    returns an error stating which key is repeated and in which file the
    duplication occurs.

    Parameters
    ----------
    src : object
        Source file object

    Raises
    ------
    ConstructorError
        If duplicated keys are found, returns an error
    """

    class PreserveDuplicatesLoader(yaml.loader.Loader):
        # We deliberately define a fresh class inside the function,
        # because add_constructor is a class method and we don't want to
        # mutate pyyaml classes.
        pass

    def map_constructor(loader, node, deep=False):
        """
        Mapping, finds any duplicate keys.
        """
        mapping = {}
        for key_node, value_node in node.value:
            key = loader.construct_object(key_node, deep=deep)
            value = loader.construct_object(value_node, deep=deep)

            if key in mapping:
                esm_parser.user_error(
                    "Duplicated variables",
                    "Key ``{0}`` is duplicated {1}\n\n".format(
                        key, str(key_node.start_mark).replace("  ", "").split(",")[0]
                    ),
                )

            mapping[key] = value

        return loader.construct_mapping(node, deep)

    PreserveDuplicatesLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, map_constructor
    )
    new_loader = create_env_loader(loader=PreserveDuplicatesLoader)
    return yaml.load(src, Loader=new_loader)


class EsmConfigFileError(Exception):
    """
    Exception for yaml file containing tabs or other syntax issues.

    An exception used when yaml.load() throws a yaml.scanner.ScannerError.
    This error occurs mainly when there are tabs inside a yaml file or
    when the syntax is incorrect. If tabs are found, this exception returns
    a user-friendly message indicating where the tabs are located in the
    yaml file.

    Parameters
    ----------
    fpath : str
        Path to the yaml file
    """

    def __init__(self, fpath, yaml_error):
        report = ""
        # Loop through the lines inside the yaml file searching for tabs
        with open(fpath) as yaml_file:
            for n, line in enumerate(yaml_file):
                # Save lines and line numbers with tabs
                if "\t" in line:
                    report += str(n) + ":" + line.replace("\t", "____") + "\n"

        # Message to return
        if len(report) == 0:
            # If no tabs are found print the original error message
            try:
                print("\n\n\n" + yaml_error)
            except:
                self.message = ("\n\n A syntax error has been found in "+fpath+"\n")
        else:
            # If tabs are found print the report
            self.message = (
                "\n\n\n"
                f"Your file {fpath} has tabs, please use ONLY spaces!\n"
                "Tabs are in following lines:\n" + report
            )
        super().__init__(self.message)



import sys
import urllib
import urllib.request
from time import sleep

import esm_parser
import esm_utilities
import yaml

import esm_tools


class MessageOfTheDayError(Exception):
    """Raise this error when Message of The Day has a problem"""


class MessageOfTheDayHandler:
    """
    Message of the day class. An object from this class loads the MOTD information
    stored in the repository, and includes the methods necessary to decide which MOTD
    to display, based on the version information.
    """

    def __init__(self):
        """
        Object initialization method in charge of loading the MOTD information from the
        repository.
        """
        url = "https://raw.githubusercontent.com/esm-tools/esm_tools/release/motd/motd.yaml"
        try:
            self.motdfile = urllib.request.urlopen(url)
        except urllib.error.HTTPError:
            timeout = 1  # seconds to wait
            # print(f"HTTP Error: Connection to file {url} containing update messages could not be established")
            # print("    Please check the URL by manually...")
            # print(f"    Program will proceed in {timeout} seconds\n")
            # sleep(timeout)
            self.database_connected = False
            self.message_dict = {}
            return
        self.database_connected = True
        self.message_dict = yaml.load(self.motdfile, Loader=yaml.FullLoader)

    def action_finder(self, action):
        """
        Interprets the actions specified in the MOTD file.

        Parameters
        ----------
        action : str
            Name of the action.

        Returns
        -------
        (action, option) : tuple
            A tuple containing the name of the action and an option.
        """
        action = action.upper()
        if action.startswith("DELAY"):
            waittime = int(
                action.replace("DELAY", "").replace("(", "").replace(")", "").strip()
            )
            return ("sleep", waittime)
        if action.startswith("ERROR"):
            return ("error", 0)

    def action_handler(self, action, time, package, version):
        """
        Executes the specified action. ``sleep`` and ``error`` actions supported.

        Parameters
        ----------
        action : str
            Name of the action to be performed (``sleep`` or ``error`` supported).
        time : int
            Number of seconds to ``sleep``.
        package : str
            Package name.
        version : str
            Version number.
        """
        if action:
            if action == "sleep":
                sleep(time)
            elif action == "error":
                esm_parser.user_error(
                    "Version",
                    (
                        f"Version {version} of '{package}' package has been tagged as "
                        + "problematic. Please, change version. For more information, "
                        + "please, read the message above."
                    ),
                )

    def motd_handler(self, mypackage, myversion):
        """
        Method to print the required MOTD messages.

        Parameters
        ----------
        mypackage : str
            Name of the ESM package to be checked. Currently only ``esm-tools`` after
            the monorepo rework.
        myversion : str
            Current version of the package in the format ``X.Y.Z``.
        """
        if not self.database_connected:
            return
        action = None
        time = -1
        # Loop through messages
        for message in self.message_dict:
            # If the package for this message is the same as ``mypackage`` and the
            # the version condition is met display the MOTD
            if self.message_dict[message][
                "package"
            ] == mypackage and esm_utilities.check_valid_version(
                self.message_dict[message]["versions"], version=myversion
            ):
                print(
                    "************************************************************************************"
                )
                print(f"Message found for package {mypackage} version {myversion}:")
                print()
                print(self.message_dict[message]["message"])
                if mypackage == "esm_tools":
                    print("Upgrade ESM-Tools by:")
                    print(f"    cd {esm_tools._get_real_dir_from_pth_file('')}")
                    print("    git pull")
                # Deprecated after monorepo rework, rewrite if we ever get to have
                # more than one package again.
                else:
                    print(
                        f"Upgrade this package by typing:              esm_versions upgrade {mypackage}"
                    )
                    print(
                        f"Upgrade all packages by typing:              esm_versions upgrade"
                    )
                print(
                    "************************************************************************************"
                )
                # Perform action
                thisaction, thistime = self.action_finder(
                    self.message_dict[message]["action"]
                )
                if thisaction == "error" or thistime > time:
                    action = thisaction
                    time = thistime
                    self.action_handler(action, time, mypackage, myversion)


def check_all_esm_packages():
    """
    Method called by ``esm_master`` and ``esm_runscripts`` (through their ``cli.py``
    files, to check for the message of the day.
    """
    # Instance the ``motd`` object
    motd = MessageOfTheDayHandler()
    # Check current version of ESM-Tools
    motd.motd_handler("esm_tools", esm_tools.__version__)


if __name__ == "__main__":
    mypackage = "esm_tools"
    myversion = "1.0.0"

    motd = MessageOfTheDayHandler()
    # Uncomment the following lines For testing using the local motd.yaml
    #import os
    #local_motd = f"{os.path.dirname(__file__)}/../../esm_tools/motd/motd.yaml"
    #print(local_motd)
    #with open(local_motd, "r") as motdfile:
    #    motd.message_dict = yaml.load(motdfile, Loader=yaml.FullLoader)
    motd.motd_handler(mypackage, myversion)
    sys.exit(0)



#!/usr/bin/env python
"""
A small wrapper that combines the shell interface and the Python interface
"""

# Import from Python Standard Library
from loguru import logger

from .info import *
from .initialization import *
from .output import *
from .read_shipped_data import *
from .repos import *
from .test_utilities import *
from .tests import *

import os
import sys


def main():
    # Logger
    logger.remove()
    logger.add(
        sys.stderr,
        filter={"": "WARNING", "esm_tests": "DEBUG"},
        format="<level>{message}</level>",
    )

    logger.info("")
    logger.info("Welcome to ESM-Tests!")
    logger.info("=====================")
    logger.info("")

    if os.environ.get("CI", False):
        logger.add(
            "out.log",
            filter={"": "WARNING", "esm_tests": "DEBUG"},
            backtrace=True,
            diagnose=True,
        )

    # Info instancing
    info = Info()

    # Parsing
    save_flag, print_state, delete_tests = info.argparse()

    # Update ``resources``
    update_resources_submodule(info)

    info["this_computer"] = (
        determine_computer_from_hostname().split("/")[-1].replace(".yaml", "")
    )
    info["last_tested_dir"] = get_last_tested_dir()

    # Predefined for later
    user_scripts = dict(comp={}, run={})

    # Print state if necessary
    if print_state:
        current_state = get_state_yaml()
        print_results(current_state, info)
        sys.exit(1)

    # Get user info for testing
    user_config(info)

    # User-specific info to remove from the files ``last_tested`` files
    info["rm_user_info"] = {
        "TEST_DIR": info["user"]["test_dir"],
        "HOME_DIR": f"{os.path.expanduser('~')}",
    }

    # Define lines to be ignored during comparison
    try:
        info["ignore"] = get_ignore_compare_yaml()
    except FileNotFoundError as e:
        print("Whoops, that did not work... I was looking here:")
        print(f"{info['script_dir']}/ignore_compare.yaml")
        for f in os.listdir(info["script_dir"]):
            print(f)
        print(e)
        raise

    # Special actions for running from GitHub servers
    if info["in_github"]:
        info["rm_user_info"]["HOME_DIR"] = "/__w/esm_tools"
        # Ignore the globbing variables
        info["ignore"]["finished_config"].append("_glob_[0-9]*: ")

    logger.debug(f"User info: {info.get('user')}")
    logger.debug(f"Actually compile: {info.get('actually_compile')}")
    logger.debug(f"Actually run: {info.get('actually_run')}")

    # Gather scripts
    info = get_scripts(info)

    # Complete scripts_info
    info = read_info_from_rs(info)

    # Delete previous test
    if delete_tests:
        del_prev_tests(info)

    # Compile
    comp_test(info)

    # Run
    run_test(info)

    # Print results
    results = format_results(info)
    print_results(results, info)

    # Check if all tests passed
    if info["system_exit_on_errors"]:
        check_perfect(info, results)

    # Save files
    if save_flag == "Not defined":
        save_files(info, False)
    elif save_flag == "true" or save_flag == "True":
        save_files(info, True)

    # yprint(info["scripts"])



"""
Module Docstring.,..?
"""
import copy
import logging
import sys


def find_remaining_minutes(seconds):
    """
    Finds the remaining full minutes of a given number of seconds

    Parameters
    ----------
    seconds : int
        The number of seconds to allocate

    Returns
    -------
    int
        The leftover seconds once new minutes have been filled.
    """
    return seconds % 60


# NOTE: This actually kills the docstring, but minutes and seconds are the
# same...
find_remaining_hours = find_remaining_minutes


def date_range(start_date, stop_date, frequency):
    if isinstance(start_date, str):
        start_date = Date(start_date)
    if isinstance(stop_date, str):
        stop_date = Date(stop_date)
    if isinstance(frequency, str):
        frequency = Date(frequency)
    current_date = start_date
    while current_date <= stop_date:
        yield current_date
        current_date += frequency


class Dateformat(object):
    datesep = ["", "-", "-", "-", " ", " ", "", "-", "", "", "/"]
    timesep = ["", ":", ":", ":", " ", ":", ":", "", "", "", ":"]
    dtsep = ["_", "_", "T", " ", " ", " ", "_", "_", "", "_", " "]

    def __init__(self, form=1, printhours=True, printminutes=True, printseconds=True):
        self.form = form
        self.printseconds = printseconds
        self.printminutes = printminutes
        self.printhours = printhours

    def __repr__(self):
        return (
            "Dateformat(form=%s, printhours=%s, printminutes=%s, printseconds=%s)"
            % (self.form, self.printhours, self.printminutes, self.printseconds)
        )


class Calendar(object):
    """
    Class to contain various types of calendars.

    Parameters
    ----------
    calendar_type : int
        The type of calendar to use.

        Supported calendar types:
        0
            no leap years
        1
            proleptic greogrian calendar (default)
        ``n``
            equal months of ``n`` days

    Attributes
    ----------
    timeunits : list of str
        A list of accepted time units.
    monthnames : list of str
        A list of valid month names, using 3 letter English abbreviation.

    Methods
    -------
    isleapyear(year)
        Returns a boolean testing if the given year is a leapyear

    day_in_year(year)
        Returns the total number of days in a given year

    day_in_month(year, month)
        Returns the total number of days in a given month for a given year
        (considering leapyears)
    """

    timeunits = ["years", "months", "days", "hours", "minutes", "seconds"]
    monthnames = [
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ]

    def __init__(self, calendar_type=1):
        self.calendar_type = calendar_type

    def isleapyear(self, year):
        """
        Checks if a year is a leapyear

        Parameters
        ----------
        year : int
            The year to check

        Returns
        -------
        bool
            True if the given year is a leapyear
        """
        if self.calendar_type == 1:
            if (year % 4) == 0:
                if (year % 100) == 0:
                    if (year % 400) == 0:
                        leapyear = True
                    else:
                        leapyear = False
                else:
                    leapyear = True
            else:
                leapyear = False
        else:
            leapyear = False
        return leapyear

    def day_in_year(self, year):
        """
        Finds total number of days in a year, considering leapyears if the
        calendar type allows for them.

        Parameters
        ----------
        year : int
            The year to check

        Returns
        -------
        int
            The total number of days for this specific calendar type
        """
        if self.calendar_type == 0:
            number_of_days = 365
        elif self.calendar_type == 1:
            number_of_days = 365 + int(self.isleapyear(year))
        else:
            number_of_days = self.calendar_type * 12
        return number_of_days

    def day_in_month(self, year, month):
        """
        Finds the number of days in a given month

        Parameters
        ----------
        year : int
            The year to check
        month : int or str
            The month number or short name.

        Returns
        -------
        int
            The number of days in this month, considering leapyears if needed.

        Raises
        ------
        TypeError
            Raised when you give an incorrect type for month
        """

        if isinstance(month, str):
            month = month.capitalize()  # Clean up possible badly formated month
            month = self.monthnames.index(month) + 1  # Remember, python is 0 indexed
        # Make sure you gave a month from 1 to 12
        if month > 12:
            raise TypeError("You have given an idiotic month, please reconsider")
        elif not isinstance(month, int):
            raise TypeError(
                "You must supply either a str with short month name, or an int!"
            )
        if self.calendar_type == 0 or self.calendar_type == 1:
            if month in [1, 3, 5, 7, 8, 10, 12]:
                return 31
            if month in [4, 6, 9, 11]:
                return 30
            return 28 + int(self.isleapyear(year))
        # I don't really like this, but if the calendar type is not 1 or 0, it
        # is ``n``, with n being the number of days in equal-length months...
        return self.calendar_type

    def __repr__(self):
        return "esm_calendar(calendar_type=%s)" % self.calendar_type

    def __str__(self):
        if self.calendar_type == 0:
            return "esm_calender object with no leap years allowed"
        if self.calendar_type == 1:
            return "esm_calendar object with allowed leap years"
        return (
            "esm_calendar object with equal-length months of %s days"
            % self.calendar_type
        )


class Date(object):
    """
    A class to contain dates, also compatiable with paleo (negative dates)

    Parameters
    ----------
    indate : str
        The date to use.

        See `pyesm.core.time_control.esm_calendar.Dateformat` for available
        formatters.

    calendar : ~`pyesm.core.time_control.esm_calendar.Calendar`, optional
        The type of calendar to use. Defaults to a greogrian proleptic calendar
        if nothing is specified.

    Attributes
    ----------
    year : int
        The year
    month : int
        The month
    day : int
        The day
    hour : int
        The hour
    minute : int
        The minute
    second : int
        The second
    _calendar : ~`pyesm.core.time_control.esm_calendar.Calendar`
        The type of calendar to use

    Methods
    -------
    """

    def __init__(self, indate, calendar=Calendar()):
        printhours = True
        printminutes = True
        printseconds = True
        ndate = ["1900", "01", "01", "00", "00", "00"]
        ds = ""
        ts = ""
        if "T" in indate:
            indate2 = indate.replace("T", "_")
            ts = ":"
        else:
            indate2 = indate
        if "_" in indate2:
            date, time = indate2.split("_")
        else:
            date = indate2
            time = ""
            ts = ":"
        for index in [3, 4, 5]:
            if len(time) == 2:
                ndate[index] = time
                time = time[2:]
            elif len(time) > 2:
                ndate[index] = time[:2]
                if len(time) > 2:
                    time = time[2:]
                    if time[0] == ":":
                        time = time[1:]
                        ts = ":"
            else:
                ndate[index] = "00"
                if index == 3:
                    printhours = False
                elif index == 4:
                    printminutes = False
                elif index == 5:
                    printseconds = False
        for index in [2, 1]:
            ndate[index] = date[-2:]
            date = date[:-2]
            if len(date) > 0 and date[-1] == "-":
                date = date[:-1]
                ds = "-"
        ndate[0] = date
        if ds == "-" and ts == ":":
            if "T" not in indate:
                form = 1
            else:
                form = 2
        elif ds == "-" and ts == "":
            form = 7
        elif ds == "" and ts == ":":
            form = 6
        elif ds == "" and ts == "":
            form = 9
        self.year, self.month, self.day, self.hour, self.minute, self.second = map(
            int, ndate
        )
        (
            self.syear,
            self.smonth,
            self.sday,
            self.shour,
            self.sminute,
            self.ssecond,
        ) = map(str, ndate)

        self._calendar = calendar
        self.doy = self.day_of_year()
        self.sdoy = str(self.day_of_year())

        self._date_format = Dateformat(form, printhours, printminutes, printseconds)

    @property
    def sdoy(self):
        return self.__sdoy

    @sdoy.setter
    def sdoy(self, sdoy):
        self.__sdoy = str(self.doy)

    @property
    def syear(self):
        return self.__syear

    @syear.setter
    def syear(self, syear):
        self.__syear = str(self.year)

    @property
    def smonth(self):
        return self.__smonth

    @smonth.setter
    def smonth(self, smonth):
        self.__smonth = str(self.month).zfill(2)

    @property
    def sday(self):
        return self.__sday

    @sday.setter
    def sday(self, sday):
        self.__sday = str(self.day).zfill(2)

    @property
    def shour(self):
        return self.__shour

    @shour.setter
    def shour(self, shour):
        self.__shour = str(self.hour).zfill(2)

    @property
    def sminute(self):
        return self.__sminute

    @sminute.setter
    def sminute(self, sminute):
        self.__sminute = str(self.minute).zfill(2)

    @property
    def ssecond(self):
        return self.__ssecond

    @ssecond.setter
    def ssecond(self, ssecond):
        self.__ssecond = str(self.second).zfill(2)

    def output(self, form="SELF"):
        return self.format(form)

    @classmethod
    def from_list(cls, _list):
        """
        Creates a new Date from a list

        Parameters
        ----------
        _list : list of ints
            A list of [year, month, day, hour, minute, second]

        Returns
        -------
        date : ~`pyesm.core.time_control.esm_calendar.Date`
            A new date of year month day, hour minute, second
        """

        negative = ["-" if _list[i] < 0 else "" for i in range(6)]
        _list = [str(abs(element)) for element in _list]

        if len(_list[0]) < 4:
            _list[0] = _list[0].zfill(4)
        _list[1:6] = [
            element.zfill(2) if len(element) < 2 else element for element in _list[1:6]
        ]

        _list = [negative[i] + _list[i] for i in range(6)]

        indate = (
            _list[0]
            + "-"
            + _list[1]
            + "-"
            + _list[2]
            + "_"
            + _list[3]
            + ":"
            + _list[4]
            + ":"
            + _list[5]
        )

        return cls(indate)

    fromlist = from_list

    def __repr__(self):
        return "Date(%s-%s-%sT%s:%s:%s)" % (
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
        )

    def __getitem__(self, item):
        return (self.year, self.month, self.day, self.hour, self.minute, self.second)[
            item
        ]

    def __setitem__(self, item, value):
        value = int(value)
        if item == 0:
            self.year = value
        elif item == 1:
            self.month = value
        elif item == 2:
            self.day = value
        elif item == 3:
            self.hour = value
        elif item == 4:
            self.minute = value
        elif item == 5:
            self.second = value
        else:
            raise IndexError("You can only assign up to 5!")

    def __lt__(self, other):
        self_tup = (
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
        )
        other_tup = (
            other.year,
            other.month,
            other.day,
            other.hour,
            other.minute,
            other.second,
        )
        return self_tup < other_tup

    def __le__(self, other):
        self_tup = (
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
        )
        other_tup = (
            other.year,
            other.month,
            other.day,
            other.hour,
            other.minute,
            other.second,
        )
        return self_tup <= other_tup

    def __eq__(self, other):
        self_tup = (
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
        )
        other_tup = (
            other.year,
            other.month,
            other.day,
            other.hour,
            other.minute,
            other.second,
        )
        return self_tup == other_tup

    def __ne__(self, other):
        self_tup = (
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
        )
        other_tup = (
            other.year,
            other.month,
            other.day,
            other.hour,
            other.minute,
            other.second,
        )
        return self_tup != other_tup

    def __ge__(self, other):
        self_tup = (
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
        )
        other_tup = (
            other.year,
            other.month,
            other.day,
            other.hour,
            other.minute,
            other.second,
        )
        return self_tup >= other_tup

    def __gt__(self, other):
        self_tup = (
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
        )
        other_tup = (
            other.year,
            other.month,
            other.day,
            other.hour,
            other.minute,
            other.second,
        )
        return self_tup > other_tup

    def __sub__(self, other):
        if isinstance(other, Date):
            return self.sub_date(other)
        elif type(other) == tuple:
            return self.sub_tuple(other)
        else:
            print("No known combination for subtraction")
            sys.exit(1)

    def sub_date(self, other):
        d2 = copy.deepcopy(
            [self.year, self.month, self.day, self.hour, self.minute, self.second]
        )
        d1 = copy.deepcopy(
            [other.year, other.month, other.day, other.hour, other.minute, other.second]
        )

        diff = [0, 0, 0, 0, 0, 0]

        while d1[1] > 1:
            diff[1] -= 1
            d1[1] -= 1
            diff[2] -= self._calendar.day_in_month(d1[0], d1[1])

        while d2[1] > 1:
            diff[1] += 1
            d2[1] -= 1
            diff[2] += self._calendar.day_in_month(d2[0], d2[1])

        while d1[2] > 1:
            diff[2] -= 1
            d1[2] -= 1

        while d2[2] > 1:
            diff[2] += 1
            d2[2] -= 1

        if diff[1] < 0:
            diff[0] = diff[0] - 1

        while d2[0] > d1[0]:
            diff[0] += 1
            diff[1] += 12
            diff[2] += self._calendar.day_in_year(d1[0])
            d1[0] += 1

        diff[3] += diff[2] * 24
        if diff[3] < 0:
            diff[3] += 24
        diff[4] += diff[3] * 60
        if diff[4] < 0:
            diff[4] += 60
        diff[5] += diff[4] * 60
        if diff[5] < 0:
            diff[5] += 60

        return diff

    def time_between(self, date, outformat="seconds"):
        """
        Computes the time between two dates

        Parameters
        ----------
        date : ~`pyesm.core.time_control.date`
            The date to compare against.

        Returns
        -------
        ??
        """
        if date > self:
            diff = date - self
        else:
            diff = self - date

        for index in range(0, 6):
            if outformat == self._calendar.timeunits[index]:
                # FIXME: Wouldn't this stop after the very first index that matches?
                # I think that is the point, but I'm not sure.
                return diff[index]
        return None  # ...? Or raise an error?

    def day_of_year(self):
        """
        Gets the day of the year, counting from Jan. 1

        Returns
        -------
        int
            The day of the current year.
        """
        if self[1] == self[2] == 1:
            return 1
        else:
            date2 = Date(str(self[0]) + "-01-01T00:00:00", self._calendar)
            return self.time_between(date2, "days") + 1

    def __str__(self):
        return (
            "-".join([str(self.year), str(self.month).zfill(2), str(self.day).zfill(2)])
            + "T"
            + ":".join(
                [
                    str(self.hour).zfill(2),
                    str(self.minute).zfill(2),
                    str(self.second).zfill(2),
                ]
            )
        )

    def format(
        self, form="SELF", givenph=None, givenpm=None, givenps=None
    ):  # basically format_date
        """
        Needs a docstring!
        The following forms are accepted:
        + SELF: uses the format which was given when constructing the date
        + 0: A Date formated as YYYY

        In [5]: test.format(form=1)
        Out[5]: '1850-01-01_00:00:00'

        In [6]: test.format(form=2)
        Out[6]: '1850-01-01T00:00:00'

        In [7]: test.format(form=3)
        Out[7]: '1850-01-01 00:00:00'

        In [8]: test.format(form=4)
        Out[8]: '1850 01 01 00 00 00'

        In [9]: test.format(form=5)
        Out[9]: '01 Jan 1850 00:00:00'

        In [10]: test.format(form=6)
        Out[10]: '18500101_00:00:00'

        In [11]: test.format(form=7)
        Out[11]: '1850-01-01_000000'

        In [12]: test.format(form=8)
        Out[12]: '18500101000000'

        In [13]: test.format(form=9)
        Out[13]: '18500101_000000'

        In [14]: test.format(form=10)
        Out[14]: '01/01/1850 00:00:00'
        """
        if form == "SELF":
            form = self._date_format.form

        ph = self._date_format.printhours
        pm = self._date_format.printminutes
        ps = self._date_format.printseconds

        if not givenph == None:
            ph = givenph
        if not givenpm == None:
            pm = givenpm
        if not givenps == None:
            ps = givenps
        ndate = list(
            map(
                str,
                (
                    self.year,
                    self.smonth,
                    self.sday,
                    self.shour,
                    self.sminute,
                    self.ssecond,
                ),
            )
        )
        if form == 0:
            if len(ndate[0]) < 4:
                for _ in range(1, 4 - len(ndate[0])):
                    ndate[0] = "0" + ndate[0]
        elif form == 5:
            temp = ndate[0]
            ndate[0] = ndate[2]
            ndate[2] = temp
            ndate[1] = self._calendar.monthnames[int(ndate[1]) - 1]
        elif form == 8:
            if len(ndate[0]) < 4:
                print("Format 8 clear with 4 digit year only")
                sys.exit(2)
        elif form == 10:
            temp = ndate[0]
            ndate[0] = ndate[1]
            ndate[1] = ndate[2]
            ndate[2] = temp

        for index in range(0, 6):
            if len(ndate[index]) < 2:
                ndate[index] = "0" + ndate[index]

        ndate[1] = self._date_format.datesep[form] + ndate[1]
        ndate[2] = self._date_format.datesep[form] + ndate[2]
        ndate[3] = self._date_format.dtsep[form] + ndate[3]
        ndate[4] = self._date_format.timesep[form] + ndate[4]
        ndate[5] = self._date_format.timesep[form] + ndate[5]

        if not ps:
            ndate[5] = ""
        if not pm and ndate[5] == "":
            ndate[4] = ""
        if not ph and ndate[4] == "":
            ndate[3] = ""

        return ndate[0] + ndate[1] + ndate[2] + ndate[3] + ndate[4] + ndate[5]

    def makesense(self, ndate):
        """
        Puts overflowed time back into the correct unit.

        When manipulating the date, it might be that you have "70 seconds", or
        something similar. Here, we put the overflowed time into the
        appropriate unit.
        """
        # ndate = copy.deepcopy(self)
        ndate[4] = ndate[4] + ndate[5] // 60
        ndate[5] = ndate[5] % 60

        ndate[3] = ndate[3] + ndate[4] // 60
        ndate[4] = ndate[4] % 60

        ndate[2] = ndate[2] + ndate[3] // 24
        ndate[3] = ndate[3] % 24

        ndate[0] = ndate[0] + (ndate[1] - 1) // 12
        ndate[1] = (ndate[1] - 1) % 12 + 1

        while ndate[2] > self._calendar.day_in_month(ndate[0], ndate[1]):
            ndate[2] = ndate[2] - self._calendar.day_in_month(ndate[0], ndate[1])
            ndate[1] = ndate[1] + 1
            ndate[0] = ndate[0] + (ndate[1] - 1) // 12
            ndate[1] = (ndate[1] - 1) % 12 + 1

        ndate[0] = ndate[0] + (ndate[1] - 1) // 12
        ndate[1] = (ndate[1] - 1) % 12 + 1

        while ndate[2] <= 0:
            ndate[1] = ndate[1] - 1
            if ndate[1] == 0:
                ndate[1] = 12
                ndate[0] = ndate[0] - 1
            ndate[2] = ndate[2] + self._calendar.day_in_month(ndate[0], ndate[1])

        if ndate[1] == 0:
            ndate[1] = 12
            ndate[0] = ndate[0] - 1

        return ndate

        # self.year, self.month, self.day, self.hour, self.minute, self.second = map(
        #    int, ndate
        # )
        # self.syear, self.smonth, self.sday, self.shour, self.sminute, self.ssecond = map(
        #    str, ndate
        # )

    def add(self, to_add):
        """
        Adds another date to this one.

        Parameters
        ----------
        to_add : ~`pyesm.core.time_control.Date`
            The other date to add to this one.

        Returns
        -------
        new_date : ~`pyesm.core.time_control.Date`
            A new date object with the added dates
        """

        me = [self.year, self.month, self.day, self.hour, self.minute, self.second]
        result = [me[i] + to_add[i] for i in range(6)]
        result = self.makesense(result)

        new_date = self.from_list(result)
        return new_date

    __add__ = add

    def sub_tuple(self, to_sub):
        """
        Adds another date to from one.

        Parameters
        ----------
        to_sub : ~`pyesm.core.time_control.Date`
            The other date to sub from this one.

        Returns
        -------
        new_date : ~`pyesm.core.time_control.Date`
            A new date object with the subtracted dates
        """

        me = [self.year, self.month, self.day, self.hour, self.minute, self.second]
        result = [me[i] - to_sub[i] for i in range(6)]
        result = self.makesense(result)

        new_date = self.from_list(result)
        return new_date



import os
import sys
import yaml

import questionary

import esm_parser
from esm_calendar import Date, Calendar


class PrevRunInfo(dict):
    """
    A dictionary subclass to access information from the previous run. The object is
    created in the ``SimulationSetup`` class in ``self.config["prev_run"]``. The idea
    behind this class is that variables from the previous run can be called from the
    yaml files with the same syntax as one would do for the current run.

    The syntax is as follows:

    .. code-block:: yaml

       <your_var>: ${prev_run.<path>.<to>.<var>}

    For example, let's assume we want to access the ``time_step`` from the previous run
    of a `FESOM` simulation and store it in a variable called ``prev_time_step``:

    .. code-block:: yaml

       prev_time_step: ${prev_run.fesom.time_step}

    .. Note:: Only the single previous simulation loaded

    .. Warning:: Use this feature only when there is no other way of accessing the
       information needed. Note that, for example, dates of the previous run are
       already available in the current run, under variables such as
       ``last_start_date``, ``parent_start_date``, etc.
    """

    def __init__(self, config={}, prev_config=None):
        """
        Links the current ``config`` and ``prev_config`` to the object.

        Parameters
        ----------
        config : dict, esm_parser.ConfigSetup
            ConfigSetup object containing the information of the current simulation.
            **Note:** this variable needs to remain untouched inside this class because
            it is not a deepCopy, and it cannot be a deepCopy because this class needs
            this config to be updated on the go to properly work.
        prev_config : dict
            Dictionary that contains information loaded from the previous config
            file/run. If not provided, it means that the object is not nested into
            another PrevRunInfo object. When provided, it means it is nested.
        """
        self._config = config
        self._prev_config = prev_config
        # Set default value of the object while the config file has not been read
        self.__setitem__("NONE_YET", {})
        # prev_run_config_file and calendar date
        self._prcf = {}
        # List of components containning variables using the ``prev_run`` feature
        self.components_with_prev_run()
        # Counter for debuggin
        self._prev_config_count = 0

    def components_with_prev_run(self):
        """
        Lists components containning variables using the ``prev_run`` feature. Reading
        of the previous config files will occur only for those components that use
        ``prev_var``.
        """
        # Search for components only if ``self._config`` is not empty
        if len(self._config) > 0:
            components = self._config.keys()
            # Make sure prev_run is not included and also that general is the last of
            # the components
            components = [
                component
                for component in components
                if component not in ["prev_run", "general"]
            ]
            components.append("general")
            # Loop through the components, and find which ones contain at least one
            # ``prev_run.`` value
            c_with_prev_run = []
            for component in components:
                if self.str_value_in_nested_dictionary(
                    "prev_run.", self._config[component]
                ):
                    c_with_prev_run.append(component)

            self._components = c_with_prev_run
        else:
            self._components = []

    def str_value_in_nested_dictionary(self, string_to_search, nested_dict):
        """
        Search recursively inside of a component for a ``string_to_search`` in the
        values of the nested dictionary ``nested_dict``.

        Parameters
        ----------
        string_to_search : str
            A string to be match in any values of the nested dictionary.
        nested_dict : dict
            Maybe, a nested dictionary which keys need to be recursively evaluated, to
            try to find in their values the ``string_to_search``. It really does not need
            to be a dictionary, as it stops the recursive search as soon as it is not a
            dictionary.

        Returns
        -------
        found : bool
            A boolean indicating if the ``string_to_search`` was found in any value.
        """
        found = False
        # If it's a dictionary, call this method recursively for each key
        if isinstance(nested_dict, dict):
            for key in nested_dict.keys():
                found = self.str_value_in_nested_dictionary(
                    string_to_search, nested_dict[key]
                )
                if found:
                    break
        # If it's a string check if ``string_to_searh`` is contained
        elif isinstance(nested_dict, str):
            if string_to_search in nested_dict:
                found = True

        return found

    def __getitem__(self, key):
        """
        Defines the special behaviour for accessing a ``key`` of the object (i.e. when
        the object is called such as ``<object>[key]``). If ``_prev_config`` is already
        loaded returns the value of the ``key``. Otherwise, it tries to load
        ``_prev_config`` and if not possible yet, returns a ``PrevRunInfo`` instance.
        """

        # If the previous config is not loaded yet (no file found), try to load it
        if not self._prev_config and len(self._config) > 0:
            self.prev_run_config()
        # If the previous config was loaded return the key
        if self._prev_config:
            if key == "NONE_YET":
                value = {}
            else:
                value = self._prev_config[key]
        # If the previous config is not loaded yet, return an instance of
        # ``PrevRunInfo``
        else:
            value = PrevRunInfo(prev_config=self._prev_config)

        return value

    def get(self, *args, **kwargs):
        """
        Defines the special behaviour for the ``get`` method of the object (i.e. when
        the object is called such as ``<object>.get(key, <value>)``). If
        ``_prev_config`` is already loaded returns the value of the ``key`` in
        ``_prev_config`` if it exists, if not it returns ``None`` if no second argument
        is defined for the ``get``, or it returns the second argument, just as a
        standard ``<dict>.get`` would do.
        """

        key = args[0]
        # If the previous config is not loaded yet (no file foudn), try to load it
        if not self._prev_config and len(self._config) > 0:
            self.prev_run_config()
        # If the previous config was loaded, use get
        if self._prev_config:
            if key == "NONE_YET":
                value = {}
            else:
                value = self._prev_config.get(*args, **kwargs)
        # If the previous config is not loaded yet, return get of an empty dict
        else:
            value = {}.get(*args, **kwargs)

        return value

    def prev_run_config(self):
        """
        If all the necessary information is available, loads the previous config file
        for each component. The component loop is only run once per model. Once loaded,
        the ``get`` and ``__get__`` methods will not call this method anymore.
        """
        # Logic for interactive questions about the config file
        fromdir = os.path.realpath(self._config["general"].get("started_from", ""))
        scriptsdir = os.path.realpath(
            f"{self._config['general'].get('experiment_dir', '')}/scripts/"
        )
        # This is necessary to display the message only once, instead of twice
        self.warn = (
            fromdir == scriptsdir
            and self._config["general"].get("jobtype", "") == "compute"
        )
        # Check for interactive, or submitted from a computing node, to avoid
        # using ``input()`` or ``questionaries`` in the second case
        self.isinteractive = self._config["general"].get("isinteractive", False)

        # Loop through components
        components = self._components
        for component in components:
            # Check if the ``prev_run_config_file`` was previously found already. If
            # not, try to find it
            if self._prcf.get(component):
                prev_run_config_file, calc_prev_date = self._prcf[component]
            else:
                prev_run_config_file, calc_prev_date = self.find_config(component)
                self._prcf[component] = (prev_run_config_file, calc_prev_date)

            # If the file exists, load the file content
            if os.path.isfile(prev_run_config_file):
                # For debugging purposes: this line should not be reached more that
                # once per component including ``prev_run`` variables
                self._prev_config_count += 1
                # print(f"PREV CONFIG COUNT: {self._prev_config_count}")

                # Open the file and load the previous run information
                with open(prev_run_config_file, "r") as prev_file:
                    prev_config = yaml.load(prev_file, Loader=yaml.FullLoader)
                # Back-compatibility with old config files
                if "dictitems" in prev_config:
                    prev_config = prev_config["dictitems"]
                # In case a ``prev_run`` info exists inside the file, remove it to
                # avoid config files from getting huge (prev_run nested inside
                # prev_run...)
                if "prev_run" in prev_config:
                    del prev_config["prev_run"]

                # Check that the data really comes from the previous run
                prev_date = prev_config["general"]["end_date"]
                prev_date_stamp = Date(prev_date).format(
                    form=9, givenph=False, givenpm=False, givenps=False
                )
                calc_prev_date_stamp = calc_prev_date.format(
                    form=9, givenph=False, givenpm=False, givenps=False
                )
                # Dates don't match
                if calc_prev_date_stamp != prev_date_stamp and self.warn:
                    esm_parser.user_note(
                        f"End date of the previous configuration file for '{component}'"
                        + " not coinciding:",
                        (
                            f"    File loaded: {prev_run_config_file}\n"
                            + f"    This previous date: {calc_prev_date}\n"
                            + f"    Previous date in prev config: {prev_date}\n"
                        ),
                    )
                    # Only ask the user about a mismatch when manually restarted
                    if self.isinteractive:
                        no_input = True
                        while no_input:
                            answer = input(f"Do you want to proceed anyway?[y/n]: ")
                            if answer == "y":
                                no_input = False
                            elif answer == "n":
                                sys.exit(0)
                            else:
                                print("Incorrect answer.")

                # Load the component info into the self._prev_config dictionary
                if not self._prev_config:
                    self._prev_config = {}
                self._prev_config[component] = prev_config[component]
                # Load the general value of this component's prev_run
                # MA: this is potentially dangerous if all the following conditions are
                # met: 1) more than one component uses the prev_run feature, 2) both
                # components are branched off and come from different spinups, meaning
                # that they don't share the same general configuration, and 3) some of
                # the models need the general configuration from the previous run.
                if component != "general":
                    self._prev_config["general"] = prev_config["general"]

    def find_config(self, component):
        """
        Finds the config file path
        (``<prev_exp_path>/config/*_finished_config.yaml_<DATE>``) required for the
        accessing the information of the previous experiment. It contemplates the
        following cases:

        a. **first spinup run:** no need for providing any path

        b. **automatic continuation of a run in the same directory:** get the path to
           the previous config file by matching with the end date. If multiple files
           match the end date, throw an error indicating the user how to proceed.

        c. **manual continuation of a run in the same directory:** get the path to
           the previous config file by matching with the end date. If multiple files
           match the end date, ask the user which one to use.

        d. **branching off experiment:** if this experiment branches off from a parent
           experiment, return the ``prev_run_config_file`` variable of the evaluated
           ``component`` as the path. If ``prev_run_config_file`` is not defined in the
           runscript, throw an error indicating how to proceed.

        Parameters
        ----------
        component : str
            Component for which the previous config file needs to be located.

        Returns
        -------
        prev_run_config_path : str
            Absolute path to the previous config file.
        prev_date : esm_calendar.Date
            Previous date calculated by substracting the time step of the ``component``
            to the current date.
        """
        prev_run_config_file = ""
        # This experiment ``config_dir``
        config_dir = (
            self._config.get("general", {}).get("experiment_dir", "") + "/config/"
        )
        expid = self._config["general"]["expid"]
        it_coupled_model_name = self._config["general"]["iterative_coupled_model"]
        # Find ``lresume`` and ``run_number`` for this component
        lresume = self._config.get(component, {}).get("lresume", False)
        run_number = self._config.get("general", {}).get("run_number", 1)
        # It's run 1
        if run_number == 1:
            # It's a branchoff experiment
            if lresume:
                # The user needs to provide the path to the config file of the previous
                # run for branchoff experiments that use the prev_run feature
                user_prev_run_config_full_path = self._config[component].get(
                    "prev_run_config_file"
                )
                if not user_prev_run_config_full_path:
                    esm_parser.user_error(
                        "'prev_run_config_file' not defined",
                        "You are trying to run a branchoff experiment that uses the "
                        + f"'prev_run' functionality for '{component}' without "
                        + "specifying the path to the previous config file. "
                        + "Please, add to your runscript the following:\n\n"
                        + f"{component}:\n"
                        + "\tprev_run_config_file: <path_to_config_file>\n\n"
                        + "Note: the path to the config file from the parent "
                        + "is '<path_to_parent_exp>/configs/*_finished_*.yaml_<DATE>'.",
                    )
                # Resolve for variables in the provided path
                if "${" in user_prev_run_config_full_path:
                    user_prev_run_config_full_path = esm_parser.find_variable(
                        [component, "prev_run_config_file"],
                        user_prev_run_config_full_path,
                        self._config,
                        [],
                        True,
                    )
                # Separate the base name from the path to the file
                user_prev_run_config_file = os.path.basename(
                    user_prev_run_config_full_path
                )
                config_dir = os.path.dirname(user_prev_run_config_full_path)

            # It's a cold start
            else:
                # There is no need of prev_run for cold starts. Do nothing
                return prev_run_config_file, ""

        # Resolve config path
        config_dir = esm_parser.find_variable(
            ["general"], config_dir, self._config, [], True
        )

        # Check for errors
        if not os.path.isdir(config_dir):
            esm_parser.user_error(
                "Config folder not existing",
                (
                    f"The config folder {config_dir} does not exist. "
                    + "The existance of this folder is a requirement for the use of the "
                    + "prev_run feature."
                ),
            )

        # Calculate previous date. This point is reached some times before it is
        # calculated in prepare.py, that's why we need the calculation here. It's only
        # use is to search for the config file time stamps and do some checks for the
        # PrevRunInfo.
        current_date = Date(self._config["general"]["current_date"])
        time_step = self._config[component].get("time_step", 1)
        try:
            time_step = int(time_step)
        except ValueError:
            time_step = 1
        prev_date = current_date - (0, 0, 0, 0, 0, time_step)

        # Calculate end date for the previous run
        prev_datestamp = prev_date.format(
            form=9, givenph=False, givenpm=False, givenps=False
        )

        # List all the config files in the config folder
        config_files = [
            cf for cf in os.listdir(config_dir) if "_finished_config.yaml" in cf
        ]
        # Select the ones ending with the correct timestamp
        potential_prev_configs = []
        for cf in config_files:
            if cf.endswith(prev_datestamp):
                potential_prev_configs.append(cf)

        # CASES FOR FINDING THE CONFIG FILE
        # ---------------------------------
        # Continuing run, not branched off, but no timestamped config files. Select the
        # one without timestamp (the case for run 1 in spinup).
        if len(potential_prev_configs) == 0 and run_number > 1:
            prev_run_config_file = (
                f"{expid}_{it_coupled_model_name}finished_config.yaml"
            )
        # Continuing run, not branched off, and one potential file. That's our file!
        elif len(potential_prev_configs) == 1 and run_number > 1:
            prev_run_config_file = potential_prev_configs[0]
        # Continuing run, too many possibilities, if interactive, ask the user,
        # otherwise, crash the simulation
        elif len(potential_prev_configs) > 1 and run_number > 1:
            if self.warn:
                if self.isinteractive:
                    text = (
                        "Using the 'prev_run' feature several valid config files were"
                        + f" found for the component '{component}'."
                    )
                    # Ask the user
                    prev_run_config_file = self.ask_about_files(
                        potential_prev_configs, component, config_dir, text
                    )
                else:
                    # Error
                    esm_parser.user_error(
                        "Too many possible config files",
                        "There is more than one config file with the same final date "
                        + "as the one required for the continuation of this experiment."
                        + " Please, resubmit the simulation, then you'll be ask about "
                        + "which file you'd like to use. This error comes from the "
                        + f"PrevRunInfo class, for the '{component}' component.",
                    )
            else:
                # If started by the user in a directory different than the experiment
                # directory, scripts, then load the first option. The resubmission will
                # ask about which file really use
                prev_run_config_file = potential_prev_configs[0]
        # Branch off, load what the user specifies in the runscript (only branchoffs
        # reach this point as run_number=1 and spinup have returned already up in this
        # method)
        elif run_number == 1:
            prev_run_config_file = user_prev_run_config_file
            prev_run_config_path = f"{config_dir}/{prev_run_config_file}"
            if not os.path.isfile(prev_run_config_path):
                esm_parser.user_error(
                    "'prev_run_config_file' incorrectly defined",
                    f"The file defined in the '{component}.prev_run_config_path' "
                    + f"({prev_run_config_path}) does not exist.",
                )

        prev_run_config_path = f"{config_dir}/{prev_run_config_file}"
        return prev_run_config_path, prev_date

    def ask_about_files(self, potential_prev_configs, component, config_dir, text):
        """
        Questionary to interactively decide which file to use as the previous config
        file, when more than one option.

        Parameters
        ----------
        potential_prev_configs : list
            List of file names to ask the user about.
        component : str
            Component to be asked about.
        config_dir : dict, esm_parser.ConfigSetup
            Dictionary containing the information about the current run configuration.
        text : str
            Text to be displayed in the questionary.
        """
        questionary.print(100 * "=")
        questionary.print(text)
        questionary.print(100 * "=")

        user_confirmed = False
        while not user_confirmed:
            response = questionary.select(
                f"Which one do you want to use?",
                choices=(
                    potential_prev_configs
                    + ["[Quit] None of the files, stop simulation now"]
                ),
            ).ask()  # returns value of selection
            if "[Quit]" in response:
                if questionary.confirm("Are you sure?").ask():
                    sys.exit(0)
            user_confirmed = questionary.confirm("Are you sure?").ask()
        return response



def prepare_environment(config):
    default_input_grid = config["general"]["experiment_couple_dir"] +"/ice.griddes"
    #print (config)
    environment_dict = {
            "PISM_TO_SOLID_EARTH": 1,
            "SOLID_EARTH_TO_PISM": 1,
            "COUPLE_DIR": config["general"]["experiment_couple_dir"],
            "RESTART_DIR_pism": config["pism"]["experiment_restart_in_dir"],
            "ice_bedrock_change_file": (
                config["general"]["experiment_couple_dir"] +
                "/bedrock_change.nc"
                ),
            "POOL_DIR_pism": config["pism"]["pool_dir"],
            "DOMAIN_pism": config["pism"]["domain"],
            "EXE_pism": config["pism"]["executable"],
            "RES_pism": config["pism"]["resolution"],
            "RUN_NUMBER_pism" : config["general"]["run_number"],
            "pism_solidearth_initialize_method": config["pism"]["solidearth_initialize_method"],
            "pism_solidearth_initialize_dummyrun_file": config["pism"].get("solidearth_initialize_dummyrun_file", ""),
            "INPUT_GRID_pism": config["pism"].get("input_grid", default_input_grid),
            "INPUT_FILE_pism": config["pism"].get("cli_input_file_pism"),
            "YR0_INI_pism": config["general"]["initial_date"].syear,
            "CURRENT_YEAR_pism": config["general"]["current_date"].syear,
            "END_YEAR_pism": config["general"]["end_date"].syear,
            "NYEAR_pism_standalone": config["general"]["nyear"],
            "latest_ex_file_pism": config["pism"]["outdata_targets"]["ex_file"],
            "RUN_DATE_STAMP": config["general"]["run_datestamp"],
            "LAST_RUN_DATE_STAMP": config["general"]["last_run_datestamp"],
            }

    if config["general"]["run_number"] > 1:
        environment_dict["latest_restart_file_pism"] = config["pism"]["restart_in_targets"]["restart"]
    print (environment_dict)
    return environment_dict









def prepare_environment(config):
    default_input_grid = config["general"]["experiment_couple_dir"] +"/ice.griddes"
    environment_dict = {
            "PISM_TO_OCEAN": 0,
            "OCEAN_TO_PISM": 1,
            "COUPLE_DIR": config["general"]["experiment_couple_dir"],
            "VERSION_pism": config["pism"]["version"].replace("github", "").replace("index", "").replace("snowflake", "")[:3],
            "POOL_DIR_pism": config["pism"]["pool_dir"],
            "CURRENT_YEAR_pism": config["general"]["current_date"].syear,
            #"RUN_NUMBER_pism": config["pism"]["run_number"],
            "END_YEAR_pism": config["general"]["end_date"].syear,
            "CHUNK_START_DATE_pism": config["general"]["chunk_start_date"],
            "CHUNK_END_DATE_pism": config["general"]["chunk_end_date"],
            "EXP_ID": config["general"]["command_line_config"]["expid"],
            "ICEBERG_DIR": config["general"]["iceberg_dir"], 
            "OUTPUT_DIR_pism": config["pism"]["experiment_outdata_dir"],
            "SPINUP_FILE_pism": config["pism"]["spinup_file"],
            "MESH_DIR_fesom": config["general"]["mesh_dir"],
            "FUNCTION_PATH": "/home/ollie/lackerma/esm_tools/runscripts/awicm-pism/coupling",
            "CHUNK_SIZE_pism_standalone": config["model2"]["chunk_size"],
            #"iter_coup_interact_method_ice2oce": "BASALSHELF_WATER_ICEBERG_MODEL",
            "DOMAIN_pism": config["pism"]["domain"],
            "RES_pism": config["pism"]["resolution"],
            "EXE_pism": config["pism"]["executable"],
            "iterative_coupling_atmosphere_pism_ablation_method": "DEBM",
            "DEBM_EXE": "/home/ollie/sxu/dEBM/dEBM/Fortran-v1.0/dEBMmain",
            "iterative_coupling_atmosphere_pism_regrid_method": "DOWNSCALE", 
            #"PISM_OCEAN_PICO_BASINS_FILE": "/home/ollie/lackerma/pool_pism/basins/antarctica.16km.nc",

            #"RESTART_DIR_pism": config["pism"]["experiment_restart_in_dir"],
            #"ice_bedrock_change_file": (
            #    config["general"]["experiment_couple_dir"] +
            #    "/bedrock_change.nc"
            #    ),
            #"DOMAIN_pism": config["pism"]["domain"],
            #"EXE_pism": config["pism"]["executable"],
            #"RES_pism": config["pism"]["resolution"],
            #"RUN_NUMBER_pism" : config["general"]["run_number"],
            #"pism_solidearth_initialize_method": config["pism"]["solidearth_initialize_method"],
            #"pism_solidearth_initialize_dummyrun_file": config["pism"].get("solidearth_initialize_dummyrun_file", ""),
            #"INPUT_GRID_pism": config["pism"].get("input_grid", default_input_grid),
            "INPUT_FILE_pism": config["pism"].get("cli_input_file_pism"),
            #"YR0_INI_pism": config["general"]["initial_date"].syear,
            #"NYEAR_pism_standalone": config["general"]["nyear"],
            #"latest_ex_file_pism": config["pism"]["outdata_targets"]["ex_file"],
            #"latest_restart_file_pism": config["pism"]["restart_in_in_work"]["restart"],
            #"restart_file_pism": config["pism"]["restart_in_in_work"]["restart"],
            #"RUN_DATE_STAMP": config["general"]["run_datestamp"],
            #"LAST_RUN_DATE_STAMP": config["general"]["last_run_datestamp"],
            }
    print (environment_dict)
    return environment_dict






